#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from magnumnp.common import timedmethod, constants
import torch

__all__ = ["ExternalField"]

class ExternalField(object):
    r"""
    External Field

    :param h: External Field
    :type h: list or tuple or :class:`Tensor` or function

    :Examples:

    .. code::

        # homogenious, constant field
        external = ExternalField([Hx, 0, 0])

        # homogenious, time-dependent field
        external = ExternalField(lambda state: [Hx*state.t, 0, 0])

        # inhomogenious, constant field
        x, y, z = SpatialCoordinate(state)
        external = ExternalField(Expression([x,y,z]))
    """
    def __init__(self, h = None):
        if h != None:
            self.__setattr__("h", h)

    @timedmethod
    def h(self, state):
        h = self._h(state)
        return state.convert_tensorfield(h)

    def __setattr__(self, name, value):
        if name == "h":
            if callable(value):
                self._h = value
            else:
                self._h = lambda state: value
        else:
            super().__setattr__(name, value)

    def E(self, state, domain = Ellipsis):
        E = - constants.mu_0 * state.material["Ms"] * state.m * self.h(state) * state.mesh.cell_volumes
        return E[domain].sum()



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import torch

__all__ = ["BC", "PlaneBC", "Plane"]

class BC:
    r"""
    Boundary conditions defined on cell centers. Used for Dirichlet boundary conditions with the ``LLGWithLESolver``.

    Examples
    --------

    1. Standard usage:
        .. code:: python

            u_mask = torch.zeros(n, dtype=torch.bool)
            u_mask[-1] = 1
            
            u_vals = torch.zeros((n[1], n[2], 3))

            dirichlet_bcs = []
            dirichlet_bcs.append(BC(u_mask, u_vals))

            state.bcs = dict()
            state.bcs["ud"] = dirichlet_bcs

    2. With callable:
        .. code:: python

            u_mask = torch.zeros(n, dtype=torch.bool)
            u_mask[0] = 1
            
            def u_cond(state):
                n = state.mesh.n

                A = 1e-11
                omega = 2*np.pi*4e9
                u0 = torch.zeros((n[1], n[2], 3))
                u0[:,:,0] = A*np.cos(omega*float(state.t))
            
            return u0

            dirichlet_bcs = []
            dirichlet_bcs.append(BC(u_mask, u_cond))

            state.bcs = dict()
            state.bcs["ud"] = dirichlet_bcs

    Parameters
    ----------
    mask : :class:`torch.Tensor`
        boolean tensor that is used to set the displacement via ``state.ud[mask] = condition`` or ``state.ud[mask] = condition(state)``.
    condition : :class:`torch.Tensor` or callable
        Boundary values given by a tensor, or a function of state that returns a tensor. The shape of ``condition`` has to match ``state.ud[mask]``
    """

    def __init__(self, mask, condition):
        self._mask = mask
        
        if callable(condition):
            self._condition = condition
        elif isinstance(condition, torch.Tensor):
            self._condition = lambda state : condition
        else:
            raise Exception("BC condition has invalid type " + type(condition) + "!/n condition has to be either torch.Tensor or a function of the state that returns torch.Tensor.")

    @property
    def mask(self):
        return self._mask
    
    def condition(self, state):
        return self._condition(state)

class PlaneBC:
    r"""
    Boundary conditions defined on cell boundaries. Used for Neumann boundary conditions with the ``LLGWithLESolver``.

    Examples
    --------

    1. Standard usage:
        .. code:: python

            nx, ny, nz = state.mesh.n
            t_cond = torch.zeros((ny, nz, 3))

            neumann_bcs = []
            neumann_bcs.append(PlaneBC(Plane(0, 0, -1), t_cond))
            
            state.bcs = dict()
            state.bcs["t"] = neumann_bcs

    2. With callable:
        .. code:: python

            def t_cond(state, t0):
                nx, ny, nz = state.mesh.n

                t_in = torch.zeros((ny, nz, 3))
                t_in[:,:,2] = t0 * np.sin(float(state.t) * omega)

            return t_in

            neumann_bcs = []
            neumann_bcs.append(PlaneBC(Plane(0, 0, -1), lambda state : t_cond(state, t0)))
            
            state.bcs = dict()
            state.bcs["t"] = neumann_bcs

    Parameters
    ----------
    plane : :class:`Plane`
        The plane on which the boundary condition is defined.
    condition : :class:`torch.Tensor` or callable
        Boundary values given by a tensor, or a function of state that returns a tensor. The shape of ``condition`` needs to match the shape of ``plane``
    """

    def __init__(self, plane, condition):
        self.plane = plane

        if callable(condition):
            self._condition = condition
        elif isinstance(condition, torch.Tensor):
            self._condition = lambda state : condition
        else:
            raise Exception("BC condition has invalid type " + type(condition) + "!/n condition has to be either torch.Tensor or a function of the state that returns torch.Tensor.")

    def condition(self, state):
        return self._condition(state)
    
class Plane:
    r"""
    Defines a plane for ``PlaneBC``

    Examples
    --------

    1. A plane on the top :math:`z` boundary, with the normal pointing outwards:
        .. code:: python

            Plane(2, -1, 1)

    2. A plane normal to :math:`y` at :math:`y=\Delta y_0 + \Delta y_1`, on the first 3 layers along :math:`z`:
        .. code:: python

           Plane(1, 2, -1, trans_lim2=[0,4])

    Parameters
    ----------
    normal_dim : int
        The direction of the normal vector (0=x, 1=y, 2=z).
    normal_position_index : int
        The cell's positional index along the normal direction
    normal_sign : int 
        For ``i=normal_position_index``, ``-1`` marks the interface to the cell with index ``i-1`` , ``1`` the interface to the cell with index ``i+1``.
    trans_lim1 : list of int
        Bounds of the plane on the transversal direction with lowest index
    trans_lim2 : list of int
        Bounds of the plane on the transversal direction with highest index
    """

    def __init__(self, 
                 normal_dim, # (int) direction of the normal vector (0=x, 1=y, 2=z)
                 normal_position_index, # (int) positional index in the normal dimension
                 normal_sign, # (int), 1 or -1, direction in which the normal vector points, points away from the interiour
                 trans_lim1=[0,None], # bounds of the transversal direction with lowest index
                 trans_lim2=[0,None]): # bounds of the transversal direction with the higgest index
        
        assert isinstance(normal_dim,int)
        self.dim = normal_dim

        assert isinstance(normal_position_index,int)
        self.pos = normal_position_index

        assert isinstance(normal_sign, int)
        if (self.pos == 0):
            assert normal_sign == -1
        if (self.pos == -1):
            assert normal_sign == 1
            
        self.sign = normal_sign

        assert isinstance(trans_lim1, list) or isinstance(trans_lim1, tuple) 
        assert isinstance(trans_lim1[0], int) or isinstance(trans_lim1[0], None)
        self.trans_slice1 = slice(trans_lim1[0], trans_lim1[1])

        assert isinstance(trans_lim2, list) or isinstance(trans_lim2, tuple) 
        assert isinstance(trans_lim2[0], int) or isinstance(trans_lim2[0], None)
        self.trans_slice2 = slice(trans_lim2[0], trans_lim2[1])

        if self.dim == 0:
            self.trans_dim1 = 1
            self.trans_dim2 = 2
        elif self.dim == 1:
            self.trans_dim1 = 0
            self.trans_dim2 = 2
        elif self.dim == 2:
            self.trans_dim1 = 0
            self.trans_dim2 = 1

        self.slices = [None,None,None]
        self.slices[self.dim] = self.pos
        self.slices[self.trans_dim1] = self.trans_slice1
        self.slices[self.trans_dim2] = self.trans_slice2
        self.slices = tuple(self.slices)

    def get_mask(self, state):
        mask = torch.zeros(state.mesh.n, dtype=torch.bool)
        mask[self.slices] = True
        return mask


#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from magnumnp.common import timedmethod, constants
import torch
from .field_terms import LinearFieldTerm

__all__ = ["AtomisticRuXExchangeField"]

class AtomisticRuXExchangeField(LinearFieldTerm):
    r"""
    Noncollinear coupling caused by Ru-X Spacer Layers:
    see: https://www.science.org/doi/full/10.1126/sciadv.abd8861
         https://journals.aps.org/prb/abstract/10.1103/PhysRevB.106.054401

    :param Jij: List of Coupling Constants
    :type Jij: list, optional
    """
    parameters = ["J"]

    def __init__(self, Jij, hom = None, **kwargs):
        self._hom = hom
        self._Jij = Jij
        super().__init__(**kwargs)

    @timedmethod
    def h(self, state):
        h = torch.zeros_like(state.m)
        J = self._Jij
        mu = state.material["Ms"]*state.mesh.cell_volumes
        mat = torch.tensor(state.material["RuxDistribution"].repeat_interleave(3).reshape(state.mesh.n + (3,)))
        Jsize = list(J.size())
        for i in range(Jsize[1]):
            f = 2. / (constants.mu_0 * mu)
            smx, smy, smz = 1., 1., 1.
            for k in range(i):
                smx *= (1 - torch.sign(mat[1+k:-i+k,:,:]))
                smy *= (1 - torch.sign(mat[:,1+k:-i+k,:]))
                smz *= (1 - torch.sign(mat[:,:,1+k:-i+k]))

            h[+i+1:,:,:,:] += f[+i+1:,:,:,:] * J[mat[+i+1:,:,:,:] + mat[:-i-1,:,:,:], i] * smx * state.m[:-i-1,:,:,:]
            h[:-i-1,:,:,:] += f[:-i-1,:,:,:] * J[mat[+i+1:,:,:,:] + mat[:-i-1,:,:,:], i] * smx * state.m[+i+1:,:,:,:]
            h[:,+i+1:,:,:] += f[:,+i+1:,:,:] * J[mat[:,+i+1:,:,:] + mat[:,:-i-1,:,:], i] * smy * state.m[:,:-i-1,:,:]
            h[:,:-i-1,:,:] += f[:,:-i-1,:,:] * J[mat[:,+i+1:,:,:] + mat[:,:-i-1,:,:], i] * smy * state.m[:,+i+1:,:,:]
            h[:,:,+i+1:,:] += f[:,:,+i+1:,:] * J[mat[:,:,+i+1:,:] + mat[:,:,:-i-1,:], i] * smz * state.m[:,:,:-i-1,:]
            h[:,:,:-i-1,:] += f[:,:,:-i-1,:] * J[mat[:,:,+i+1:,:] + mat[:,:,:-i-1,:], i] * smz * state.m[:,:,+i+1:,:]

        if self._hom is not None:
            for i in self._hom:
                h[i[0]:i[1],i[2]:i[3],i[4]:i[5],0] = torch.mean(h[i[0]:i[1],i[2]:i[3],i[4]:i[5], 0])
                h[i[0]:i[1],i[2]:i[3],i[4]:i[5],1] = torch.mean(h[i[0]:i[1],i[2]:i[3],i[4]:i[5], 1])
                h[i[0]:i[1],i[2]:i[3],i[4]:i[5],2] = torch.mean(h[i[0]:i[1],i[2]:i[3],i[4]:i[5], 2])
        return h



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from magnumnp.common import logging, timedmethod, constants
import torch

__all__ = ["SD_solver"]

class SD_solver(object):
    def __init__(self, terms, dm_max = 1e2, samples = 10):
        """
        This class implements the direct energy minimizing algorithm introduced in [Exl2014]_.

        .. note:: This feature is experimental.

        *Example*
          .. code:: python

            state = State(mesh)
            minimizer = Minimizer_BB([ExchangeField()])
            minimizer.minimize(state)

        *Arguments*
          terms ([:class:`LLGTerm`])
            List of LLG contributions to be considered for energy minimization
          region (:class:`str`)
            region on which the energy is minimized
          tau_min (:class:`float`)
            minimum step size
          tau_max (:class:`float`)
            maximum step size
          dm_max (:class:`float`)
            stop criterion given as supremum norm of dm/dt
          sample (:class:`int`)
            number of subsequent steps the stop criterion has to be fulfilled
        """
        self._terms = terms
        self._dm_max = dm_max
        self._samples = samples

    def _dm(self, state):
        h = sum([term.h(state) for term in self._terms])
        return torch.linalg.cross(state.m, torch.linalg.cross(state.m, h))
    
    def minimize(self, state):
        step = 0
        dm_max = 1e18
        last_dm_max = []
        energy = []
        steps = []
        tau = 1e-7

        while len(last_dm_max) < self._samples or max(last_dm_max) > self._dm_max:
            h = sum([term.h(state) for term in self._terms])
            dm = torch.linalg.cross(state.m, torch.linalg.cross(state.m, h))

            m_next = state.m - tau*dm

            # update state
            state.m = m_next
            
            dm_max = dm.max()
            last_dm_max.append(dm_max)
            if len(last_dm_max) > self._samples: last_dm_max.pop(0)

            E = sum([term.E(state) for term in self._terms])
            
            logging.info_blue("Tau: %.5g, dm_max: %.5g, E: %.5g" % (tau, dm_max, E))

            # increase step count
            step += 1
            energy.append(E)
            steps.append(step)
        return state, energy, steps

    




#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import torch
from magnumnp.common import logging

__all__ = ["RKF56"]

#Runge-Kutta-Fehlberg method with stepsize control
class RKF56(object):
    def __init__(self, f, dt = 1e-15, atol = 1e-5, rtol = 1e-5):
        self._f = f
        self._dt = dt
        self._order = 5

        # Numerical Recipies 3rd Edition suggests these values:
        self._headroom = 0.9
        self._maxstep = 1e-10
        self._minscale = 0.2
        self._maxscale = 10.
        self._atol = atol
        self._rtol = rtol # currently rtol is ignored since m is normalized!
        logging.info_green("[LLGSolver] using RKF56 solver (atol = %g)" % atol)

    def _f_wrapper(self, state, t, m, **llg_args):
        t0 = state.t
        m0 = state.m.detach()
        state.t = t
        state.m = m
        f = self._f(state, **llg_args)
        state.t = t0
        state.m = m0.normalize()
        return f

    def _try_step(self, state, **llg_args):
        f, m, t, dt = self._f_wrapper, state.m, state.t, self._dt
        state._dt = dt  # update current dt in state used by thermal field class
        k1 = dt * f(state, t,               m, **llg_args)
        k2 = dt * f(state, t +  1./ 6.*dt,  m +    1. / 6.*k1, **llg_args)
        k3 = dt * f(state, t +  4./ 15.*dt, m +    4. / 75.*k1    + 16./75.*k2, **llg_args)
        k4 = dt * f(state, t +  2./ 3.*dt,  m +    5. / 6.*k1     - 8./3.*k2    + 5./2.*k3, **llg_args)
        k5 = dt * f(state, t +  4./ 5.*dt,  m -    8. / 5.*k1     + 144./25.*k2 - 4.*k3        + 16./25.*k4, **llg_args)
        k6 = dt * f(state, t +  1.*dt,      m +    361. / 320.*k1 - 18/5.*k2    + 407./128.*k3 - 11./80.*k4  + 55./128.*k5, **llg_args)
        k7 = dt * f(state, t,               m -    11. / 640.*k1                + 11./256.*k3  - 11./160.*k4 + 11./256.*k5, **llg_args)
        k8 = dt * f(state, t +  1.*dt,      m +    93. / 640.*k1  - 18./5.*k2   + 803./256.*k3 - 11./160.*k4 + 99./256.*k5 + 1.*k7, **llg_args)

        dm =             31./384.*k1 + 1125./2816.*k3 + 9./32.*k4 + 125./768.*k5 + 5./66.*k6
        rk_error = dm - (7./1408.*k1 + 1125./2816.*k3 + 9./32.*k4 + 125./768.*k5             + 5./66.*k7 + 5./66.*k8)
        return m+dm, t+dt, rk_error

    def _optimal_stepsize(self, rk_error, atol):
        norm = torch.linalg.norm(rk_error._base.flatten() / atol, torch.inf)
        if torch.isnan(norm):
            raise RuntimeError("Unexpected error norm= %.5g!" % norm)

        if norm > 1.1:
            # decrease step, no more than factor of 5, but a fraction S more
            # than scaling suggests (for better accuracy)
            r = self._headroom / torch.pow(norm, 1.0/self._order)
            if (r < self._minscale):
                r = self._minscale
        elif norm < 0.5:
            # increase step, but no more than by a factor of 5
            r = self._headroom / torch.pow(norm, 1.0/(self._order+1.0));
            if r > self._maxscale: # increase no more than factor of 5
                r = self._maxscale
            if r < 1.: # don't allow any decrease caused by S<1
                r = 1.
        else: # no change
            r = 1.

        dt_opt = self._dt * r
        if dt_opt > self._maxstep:
            dt_opt = self._maxstep
        return dt_opt

    def step(self, state, dt, rtol = None, atol = None, **llg_args):
        t0, t1 = state.t, state.t + dt
        while state.t < t1:
            _m1, _t1, err = self._try_step(state, **llg_args)
            dt_opt = torch.tensor(self._optimal_stepsize(err, atol or self._atol))
            if self._dt > dt_opt or self._dt > t1 - state.t:
                # step size was too large, retry with optimal stepsize
                # also rescale the thermal field accordingly
                self._dt = torch.min(dt_opt, t1 - state.t).detach()
                logging.debug("REVERT step: %g, new step size: %g, time: %g" % (self._dt, dt_opt, state.t))
            else:
                # accept step, adapt stepsize for next step
                state.m = _m1
                state.t = _t1
                logging.debug("ACCEPT step: %g, new step size: %g, time: %g" % (self._dt, dt_opt, state.t))
                self._dt = dt_opt
                state._step += 1



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from magnumnp.common import logging, timedmethod, constants, Timer, complex_dtype
from .field_terms import LinearFieldTerm
import numpy as np
import torch
import torch.fft
from torch import asinh, atan, sqrt, log, abs, pi
from time import time
import os

__all__ = ["VectorPotential"]

def v(x, y, z):
    x2, y2, z2 = x**2, y**2, z**2
    r = sqrt(x2 + y2 + z2)
    return y * z * log(x + r) - x2 / 2. * atan((y*z) / (x*r))

def V(x, y, z):
    return v(x,y,z) + v(y,z,x) + v(z,x,y)

def V0(x, y, z, dz):
    return V(x, y,  dz/2.-z) \
         - V(x, y, -dz/2.-z)

def V1(x, y, z, dy, dz):
    return V0(x,  dy/2.-y, z, dz) \
         - V0(x, -dy/2.-y, z, dz)

def seidov(x, y, z, dx, dy, dz):
    res = V1( dx/2.-x, y, z, dy, dz) \
        - V1(-dx/2.-x, y, z, dy, dz)
    return res / (4*pi)

def vector_far(x, y, z, dx, dy, dz):
    res = (1./sqrt(x**2 + y**2 + z**2)).nan_to_num(posinf=0, neginf=0)
    return res * dx*dy*dz / (4.*np.pi)

def vector_func(x, y, z, dx, dy, dz, p):
    res = vector_far(x, y, z, dx, dy, dz)
    near = (x**2 + y**2 + z**2) / (dx**2 + dy**2 + dz**2) < p**2
    res[near] = seidov(x[near], y[near], z[near], dx, dy, dz)
    return res


class VectorPotential(object):
    r"""
    Vector Potential originating from current density state.j

    .. math::

        \vec{A}(\vec{x}) = \frac{1}{4 \pi} \int \vec{j}(\vec{x}') \; \frac{1}{\vert \vec{x}-\vec{x}'\vert} \, d\vec{x}'.

    Intergrals over a cuboid source region are given in [seidov], and the occuring convolution can be efficiently calculated by means of an FFT method. The integrals are evaluated for target points in the centers of each cell.

    :param p: number of next neighbors for near field via Seidov's equations (default = 20)
    :type p: int, optional
    """
    def __init__(self, p = 20, cache_dir = None):
        self._p = p
        self._cache_dir = cache_dir

    def _init_A_component(self, state):
        dx = np.array(state.mesh.dx)

        shape = [1 if n==1 else 2*n for n in state.mesh.n]
        ij = [torch.fft.fftfreq(n,1/n) for n in shape] # local indices
        ij = torch.meshgrid(*ij,indexing='ij')
        x, y, z = [ij[ind]*dx[ind] for ind in range(3)]

        Ac = vector_func(x, y, z, *dx, self._p) # TODO: handle PBCs and non-equidistant grids

        dim = [i for i in range(3) if state.mesh.n[i] > 1]
        if len(dim) > 0:
            Ac = torch.fft.rfftn(Ac, dim = dim)

        return Ac #.real.clone() ## TODO: check symmetry


    def _init_A(self, state):
        if state.mesh.pbc[0] != 0 or state.mesh.pbc[1] != 0 or state.mesh.pbc[2] != 0:
            logging.warning(f"[VectorPotential]: PBCs are not used by VectorPotential! (mesh.pbc = %s)" % str(state.mesh.pbc))

        name = "/A_%s.pt" % str(state.mesh).replace(" ","")
        if self._cache_dir != None and os.path.isfile(self._cache_dir + name):
            Axx = torch.load(self._cache_dir + name, map_location=state.device)
            logging.info("[DEMAG]: Use cached VectorPotential kernel from '%s'" % (self._cache_dir + name))
        else:
            dtype = torch.get_default_dtype()
            torch.set_default_dtype(torch.float64) # always use double precision
            time_kernel = time()

            Axx = self._init_A_component(state).to(dtype=complex_dtype[dtype])

            logging.info(f"[DEMAG]: Time calculation of VectorPotential kernel = {time() - time_kernel} s")
            torch.set_default_dtype(dtype) # restore dtype

            # cache VectorPotential tensor
            if self._cache_dir != None:
                if not os.path.isdir(self._cache_dir):
                    os.makedirs(self._cache_dir)
                torch.save(Axx, self._cache_dir + name)
                logging.info("[DEMAG]: Save VectorPotential kernel to '%s'" % (self._cache_dir + name))

        return Axx

    @timedmethod
    def A(self, state):
        if not hasattr(self, "_Axx"):
            self._Axx = self._init_A(state)

        dim = [i for i in range(3) if state.mesh.n[i] > 1]
        s = [2*state.mesh.n[i] for i in range(3) if state.mesh.n[i] > 1]

        Ax = torch.zeros_like(self._Axx)
        Ay = torch.zeros_like(self._Axx)
        Az = torch.zeros_like(self._Axx)

        j = state.j # state calls j(state) if j is a function
        Ai = [torch.fft.irfftn(self._Axx * torch.fft.rfftn(j[:,:,:,ax], dim=dim, s=s), dim=dim) for ax in range(3)]

        return torch.stack([Ai[0][:state.mesh.n[0],:state.mesh.n[1],:state.mesh.n[2]],
                            Ai[1][:state.mesh.n[0],:state.mesh.n[1],:state.mesh.n[2]],
                            Ai[2][:state.mesh.n[0],:state.mesh.n[1],:state.mesh.n[2]]], dim=3)



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import torch
import os
from collections.abc import Iterable
from magnumnp.loggers import ScalarLogger, FieldLogger
from magnumnp.common import logging

__all__ = ["Logger"]

class Logger(object):
    """
    Combined Scalar- and Field-Logger class

    *Arguments*
        directory (:class:`str`)
            The name of the log file
        scalars ([:class:`str` | :class:`function`])
            The columns to be written to the log file
        scalars_every (:class:`int`)
            Write scalar to log file every nth call
        fields ([:class:`str` | :class:`function`])
            The columns to be written to the log file
        fields_every (:class:`int`)
            Write fields to log file every nth call

    *Example*
        .. code-block:: python

            # provide key strings with are available in state
            logger = Logger('data', ['m', demag.h], ['m'], fields_every = 100)

            # Actually log fields
            state = State(mesh)
            logger << state
    """
    def __init__(self, directory, scalars = [], fields = [], scalars_every = 1, fields_every = 1):
        self.loggers = {}
        self._resume_time = None
        if len(scalars) > 0:
            self.loggers["scalars"] = ScalarLogger(os.path.join(directory, "log.dat"), scalars, every = scalars_every)
        if len(fields) > 0:
            self.loggers["fields"] = FieldLogger(os.path.join(directory, "fields.pvd"), fields, every = fields_every)

    def log(self, state):
        if state.t == self._resume_time: # avoid logging directly after resume
            self._resume_time = None
            return
        for logger in self.loggers.values():
            logger << state

    def __lshift__(self, state):
        self.log(state)

    def resume(self, state):
        """
        Tries to resume from existing log files. If resume is possible
        the state object is updated with latest possible values of t
        and m and the different log files are aligned and resumed.
        If resume is not possible the state is not modiefied and the
        simulations starts from the beginning.

        *Arguments*
            state (:class:`State`)
                The state to be resumed from the log data
        """
        last_recorded_step = self.loggers["fields"].last_recorded_step()
        if last_recorded_step is None:
            logging.warning("Resume not possible. Start over.")
            return

        resumable_step = min(map(lambda logger: logger.resumable_step(), self.loggers.values()))
        assert resumable_step >= last_recorded_step + 1

        state.m, state.t = self.loggers["fields"].step_data(last_recorded_step, "m")
        logging.info_green("Resuming from step %d (t = %g)." % (last_recorded_step, state.t))

        for logger in self.loggers.values():
            logger.resume(last_recorded_step + 1)

        self._resume_time = state.t

    def is_resumable(self):
        """
        Returns True if logger can resume from log files.

        *Returns*
            :class:`bool`
                True if resumable, False otherwise
        """
        return self.loggers["fields"].last_recorded_step() is not None



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from magnumnp.common import timedmethod, constants
import torch
from .field_terms import FieldTerm

__all__ = ["ThermalField"]

class ThermalField(FieldTerm):
    r"""
    """
    parameters = ["T"]
    def __init__(self, domain=None, **kwargs):
        self._step = None
        super().__init__(**kwargs)

    @timedmethod
    #@torch.compile
    def h(self, state):
        if state._step != self._step: # update random field
            #self._sigma = torch.normal(0., 1., size = state.m.shape) # creates tensor on CPU by default
            self._sigma = torch.normal(0., 1., size = state.m.shape).to(state.device)
            self._step = state._step

        h = self._sigma * torch.sqrt(2. * state.material["alpha"]  * constants.kb * state.T / (constants.mu_0 * state.material["Ms"] * constants.gamma * state.mesh.cell_volumes * state._dt))
        return h.nan_to_num(posinf=0, neginf=0)



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from magnumnp.common import timedmethod, constants
from .field_terms import LinearFieldTerm
from math import pi
import torch

__all__ = ["DemagFieldPBC"]

class DemagFieldPBC(LinearFieldTerm):
    @timedmethod
    def h(self, state):
        m_fft = torch.fft.fftn(state.material["Ms"] * state.m, dim = [i for i in range(3) if state.mesh.n[i] > 1]).squeeze(-1) #TODO: use rfftn -> kz should be size N//2+1
        dx, dy, dz = state.mesh.dx

        kx = (2. * pi * torch.arange(state.mesh.n[0]) / state.mesh.n[0]).reshape(-1,1,1)
        ky = (2. * pi * torch.arange(state.mesh.n[1]) / state.mesh.n[1]).reshape(1,-1,1)
        kz = (2. * pi * torch.arange(state.mesh.n[2]) / state.mesh.n[2]).reshape(1,1,-1)

        div_fft = (1.-torch.exp(-1j*kx)) * m_fft[:,:,:,0] / dx \
                + (1.-torch.exp(-1j*ky)) * m_fft[:,:,:,1] / dy \
                + (1.-torch.exp(-1j*kz)) * m_fft[:,:,:,2] / dz

        u_fft = -div_fft / (4./dx**2*torch.sin(kx/2.)**2 + \
                            4./dy**2*torch.sin(ky/2.)**2 + \
                            4./dz**2*torch.sin(kz/2.)**2)
        u_fft[0,0,0] = 0

        h_fft = torch.empty_like(m_fft)
        h_fft[:,:,:,0] = (1.-torch.exp(1j*kx)) * u_fft / dx
        h_fft[:,:,:,1] = (1.-torch.exp(1j*ky)) * u_fft / dy
        h_fft[:,:,:,2] = (1.-torch.exp(1j*kz)) * u_fft / dz

        h = torch.fft.ifftn(h_fft, dim = [i for i in range(3) if state.mesh.n[i] > 1])
        return h.real

    def E(self, state, domain = Ellipsis): # TODO: remove as soon as @compile works for DemagField
        E = -0.5 * constants.mu_0 * state.material["Ms"] * state.m * self.h(state) * state.mesh.cell_volumes
        return E[domain].sum()



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import numpy as np

__all__ = ["C_cubic", "C_isotropic", "C_trigonal", "RotateC"]

def C_cubic(C11, C12, C44):
    C = [[C11, C12, C12,   0,   0,   0],
         [C12, C11, C12,   0,   0,   0],
         [C12, C12, C11,   0,   0,   0],
         [  0,   0,   0, C44,   0,   0],
         [  0,   0,   0,   0, C44,   0],
         [  0,   0,   0,   0,   0, C44]]

    return np.array(C)
    
def C_isotropic(E, nu):
    C11 = E*(1.-nu)/((1.+nu)*(1.-2.*nu))
    C12 = E*nu/((1.+nu)*(1.-2.*nu))
    C44 = 0.5*(C11 - C12)

    return C_cubic(C11, C12, C44)

def C_trigonal(C11, C12, C13, C14, C33, C44):
    C = [[C11,  C12,  C13,  C14,   0,   0],
         [C12,  C11,  C13, -C14,   0,   0],
         [C13,  C13,  C33,    0,   0,   0],
         [C14, -C14,    0,  C44,   0,   0],
         [  0,    0,    0,    0, C44, C14],
         [  0,    0,    0,    0, C14, 0.5*(C11-C12)]]

    return np.array(C)

def RotateC(C, axis, angle):
    R = RMat(axis, angle)

    map = np.array([[0,5,4],[5,1,3],[4,3,2]])

    Ct = np.zeros((3,3,3,3))
    Cr = np.zeros((3,3,3,3))
    for i in range(3):
        for j in range(3):
            for k in range(3):
                for l in range(3):
                    Ct[i,j,k,l] = C[map[i,j],map[k,l]]

    for i in range(3):
        for j in range(3):
            for k in range(3):
                for l in range(3):
                    Cr[i,j,k,l] += np.dot(R[l], Ct[i,j,k,:])
    Ct = Cr
    Cr = np.zeros((3,3,3,3))

    for i in range(3):
        for j in range(3):
            for k in range(3):
                for l in range(3):
                    Cr[i,j,k,l] += np.dot(R[k], Ct[i,j,:,l])
    Ct = Cr
    Cr = np.zeros((3,3,3,3))

    for i in range(3):
        for j in range(3):
            for k in range(3):
                for l in range(3):
                    Cr[i,j,k,l] += np.dot(R[j], Ct[i,:,k,l])
    Ct = Cr 
    Cr = np.zeros((3,3,3,3))

    for i in range(3):
        for j in range(3):
            for k in range(3):
                for l in range(3):
                    Cr[i,j,k,l] += np.dot(R[i], Ct[:,j,k,l])
    
    res = np.zeros((6,6))
    for i in range(3):
        for j in range(3):
            for k in range(3):
                for l in range(3):
                    res[map[i,j],map[k,l]] = Cr[i,j,k,l]

    return res

def cross_product_matrix(u):
    ucm = np.zeros((3,3))
    ucm[0,1] = -u[2]
    ucm[0,2] = u[1]
    
    ucm[1,0] = u[2]
    ucm[1,2] = -u[0]
    
    ucm[2,0] = -u[1]
    ucm[2,1] = u[0]

    return ucm

def RMat(axis, angle):
    if axis == "x":
        u = np.array([1,0,0])
    elif axis == "y":
        u = np.array([0,1,0])
    elif axis == "z":
        u = np.array([0,0,1])
    else:
        u = axis

    ucm = cross_product_matrix(u)
    uom = np.outer(u,u)

    R = np.cos(angle)*np.eye(3) 
    R += np.sin(angle)*ucm 
    R += (1.-np.cos(angle))*uom

    return R


#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import torch
import sys
import subprocess
from magnumnp.common import logging, Material
from magnumnp.common.io import write_vti, write_vtr

__all__ = ["complex_dtype", "normalize", "randM", "Expression"]


complex_dtype = {
    torch.float: torch.complex,
    torch.float32: torch.complex64,
    torch.float64: torch.complex128
    }


def normalize(data):
    r"""
    Helper function to normalize vectorial data inplace
    """
    data /= torch.linalg.norm(data, dim = -1, keepdim = True)
    data[...] = torch.nan_to_num(data, posinf=0, neginf=0)
    return data

def randM(data):
    r"""
    Helper function to generate uniform distibution on the unit-sphere
    """
    if data.shape[-1] != 3:
        raise ValueError("Input tensor's last dimension needs to be 3 (shape='%s')" % str(data.shape))

    theta = 2.*torch.pi*torch.rand(data.shape[:-1])
    phi = torch.acos(2.*torch.rand(data.shape[:-1])-1.)

    data[...,0] = torch.sin(phi) * torch.cos(theta)
    data[...,1] = torch.sin(phi) * torch.sin(theta)
    data[...,2] = torch.cos(phi)
    return data


def Expression(comps):
    r"""
    Helper function to create scalar or vector fields by stacking the corresponding components.

    :param comps: list of components or single torch.Tensor
    :type comps: list, :class:`Tensor`

    :Examples:

    .. code::
        n  = (1, 1, 10)
        dx = (2e-9, 2e-9, 2e-9)
        mesh = Mesh(n, dx)
        x, y, z = mesh.SpatialCoordinate()
        
        Ms = Expression(x)            # create scalar field
        Ku_axis = Expression((x,y,z)) # create vector field
    """
    if isinstance(comps, torch.Tensor):
        comps = [comps]
    return torch.stack(comps, dim=-1)


def get_gpu_with_least_memory():
    code = """
           import torch
           def get_gpu_with_least_memory():
               if not torch.cuda.is_available():
                   print(-1)
                   return
               num_gpus = torch.cuda.device_count()
               if num_gpus == 1:
                   print(0)
                   return
               gpu_memory = [torch.cuda.mem_get_info(i)[0] for i in range(num_gpus)]
               print(gpu_memory.index(max(gpu_memory)))
           get_gpu_with_least_memory()
           """

    # unindent code
    lines = code.splitlines()
    non_empty = [line for line in lines if line.strip()]
    min_indent = min((len(line) - len(line.lstrip()) for line in non_empty), default=0)
    code = "\n".join(line[min_indent:] for line in lines)

    # run in seperate process in order to prevent nvidia-smi show the process on all GPUs
    result = subprocess.check_output([sys.executable, "-c", code])
    return int(result.decode().strip())



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from magnumnp.common import timedmethod, constants
from magnumnp.field_terms.field_terms import LinearFieldTerm
from magnumnp.linear_elasticity import epsilon, epsilon_m, sigma
import torch

__all__ = ["LinearMagnetoElasticField", "MagnetoElasticField"]

class MagnetoElasticField(object):
    r"""
    Magnetoelastic Field

    This field term is obtained from the magnetoelastic energy:

    .. math::
        E = \int \left( \frac{1}{2}\epsilon^m :C:\epsilon^m - \epsilon :C:\epsilon^m \right) \text{d}\mathbf{x}

    References:
    - C.Y. Liang et al., Nanotechnology 25 (2014) 435701 (10pp), doi:10.1088/0957-4484/25/43/435701
    - Y.C. Shu et al., Mechanics of Materials 36 (2004) 975-997, doi:10.1016/j.mechmat.2003.04.004

    :param ud: Displacement. If neither this nor 'mechanical_strain' is set, `state.ud` is used.
    :type ud: :class:`torch.Tensor` or function, optional
    :param mechanical_strain: Strain in Voigt notation (mesh.n + (6,))
    :type mechanical_strain: :class:`torch.Tensor` or function, optional

    The parameters 'ud' and 'mechanical_strain' are mutually exclusive.
    """

    def __init__(self, ud = None, mechanical_strain = None, **kwargs):
        if (ud != None) and (mechanical_strain != None):
            raise Exception("[MagnetoElasticField] received arguments for both 'ud' and 'mechanical_strain', but these are mutually exclusive.")

        if ud != None:
            if callable(ud):
                self._get_ud = ud
            else:
                self._get_ud = lambda state : ud
        else:
            self._get_ud = self._default_get_ud

        if mechanical_strain != None:
            if callable(mechanical_strain):
                self._get_strain = mechanical_strain 
            else:
                self._get_strain = lambda state : mechanical_strain
        else:
            self._get_strain = self._default_get_strain

    def _default_get_ud(self, state):
        return state.ud
    
    def _default_get_strain(self, state):
        ud = self._get_ud(state)
        return epsilon(state, ud)

    @timedmethod
    @torch.compile
    def h(self, state):     
        Ms = state.material["Ms"]
        lambda_100 = state.material["lambda_100"][:,:,:,0]
        lambda_111 = state.material["lambda_111"][:,:,:,0]

        m = state.m
   
        eps = self._get_strain(state)
        eps_m = epsilon_m(state, m)
        sig_el = sigma(state, eps-eps_m)

        # factor 1/2 is dropped due to second derivative
        dx_eps_xx = -3.*lambda_100 * m[...,0]
        dy_eps_yy = -3.*lambda_100 * m[...,1]
        dz_eps_zz = -3.*lambda_100 * m[...,2]

        # factor 1/2 is already dropped by conversion to Voigt
        d_eps_x = -3.*lambda_111 * m[...,0]
        d_eps_y = -3.*lambda_111 * m[...,1]
        d_eps_z = -3.*lambda_111 * m[...,2]

        # factor 1/2 of the energy is droped due to product rule
        h = torch.zeros(state.mesh.n+(3,))
        h[:,:,:,0] = sig_el[:,:,:,0]*dx_eps_xx + sig_el[:,:,:,5]*d_eps_y + sig_el[:,:,:,4]*d_eps_z 
        h[:,:,:,1] = sig_el[:,:,:,1]*dy_eps_yy + sig_el[:,:,:,5]*d_eps_x + sig_el[:,:,:,3]*d_eps_z
        h[:,:,:,2] = sig_el[:,:,:,2]*dz_eps_zz + sig_el[:,:,:,4]*d_eps_x + sig_el[:,:,:,3]*d_eps_y

        h *= -1. / (constants.mu_0 * Ms)
        return h.nan_to_num(posinf=0, neginf=0)
    
    @torch.compile
    def E(self, state, domain = Ellipsis):
        m = state.m

        eps_m = epsilon_m(state, m)
        sig_m = sigma(state, eps_m)
        eps = self._get_strain(state)

        zeta = (0.5*eps_m - eps)*sig_m

        E =  zeta * state.mesh.cell_volumes
        return E[domain].sum()
    
class LinearMagnetoElasticField(LinearFieldTerm):
    r"""
    Magnetoelastic Field

    This field term is obtained only from the terms in the magnetoelastic energy
    that are quadratic in :math:`\vec{m}`:

    .. math::
        E = -\int \epsilon :C:\epsilon^m \text{d}\mathbf{x}

    :param ud: Displacement. If neither this nor 'mechanical_strain' is set, `state.ud` is used.
    :type ud: :class:`torch.Tensor` or function, optional
    :param mechanical_strain: Strain in Voigt notation (mesh.n + (6,))
    :type mechanical_strain: :class:`torch.Tensor` or function, optional
    """

    def __init__(self, ud = None, mechanical_strain = None, **kwargs):
        if (ud != None) and (mechanical_strain != None):
            raise Exception("[LinearMagnetoElasticField] received arguments for both 'ud' and 'mechanical_strain', but these are mutually exclusive.")

        if ud != None:
            if callable(ud):
                self._get_ud = ud
            else:
                self._get_ud = lambda state : ud
        else:
            self._get_ud = self._default_get_ud

        if mechanical_strain != None:
            if callable(mechanical_strain):
                self._get_strain = mechanical_strain 
            else:
                self._get_strain = lambda state : mechanical_strain
        else:
            self._get_strain = self._default_get_strain

    def _default_get_ud(self, state):
        return state.ud
    
    def _default_get_strain(self, state):
        ud = self._get_ud(state)
        return epsilon(state, ud)

    @timedmethod
    @torch.compile
    def h(self, state):     
        Ms = state.material["Ms"]
        lambda_100 = state.material["lambda_100"][...,0]
        lambda_111 = state.material["lambda_111"][...,0]
        C = state.material["C"]

        B1 = -3.*lambda_100*(C[...,0,0]-C[...,0,1]) / 2.
        B2 = -3.*lambda_111*C[...,3,3]

        m = state.m
        eps = self._get_strain(state)

        h = torch.zeros(state.mesh.n+(3,))
        # Note: eps[...,3:] already includes a factor of 2 due to Voigth notation
        h[...,0] = 2*B1*m[...,0]*eps[...,0] + B2*(eps[...,5]*m[...,1] + eps[...,4]*m[...,2])
        h[...,1] = 2*B1*m[...,1]*eps[...,1] + B2*(eps[...,5]*m[...,0] + eps[...,3]*m[...,2])
        h[...,2] = 2*B1*m[...,2]*eps[...,2] + B2*(eps[...,4]*m[...,0] + eps[...,3]*m[...,1])

        h *= -1. / (constants.mu_0 * Ms)
        return h.nan_to_num(posinf=0, neginf=0)



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import torch
from magnumnp.common import logging
from scipy.spatial import KDTree

__all__ = ["Voronoi"]

class Voronoi(object):
    def __init__(self, mesh, num_grains=None, seed_points=None):
        """
        Creates a Voronoi Tesselation with a given number of seed points.
        The seed points will be randomly selected with a uniform distribution.
        Optional it is possible to manually provide seed_points.

        *Example*
            .. code:: python

                # create simple Voronoi tesselation
                voi = Voronoi(mesh, 10)
                state.write_vtk(voi.domains, "data/domains.vti")

                # perform Llloyd's iteration to improve tesselation
                domains = voi.relax()
                state.write_vtk(domains, "data/domains.vti")

                # add an intergrain phase
                voi.add_intergrain_phase(2)
                state.write_vtk(voi.domains, "domains.vti")

                # set parameters material 
                Ms = 8e5
                Ms_values = torch.normal(Ms, 0.1*Ms, (11,))
                Ms_values[-1] = 0. # set Ms=0 for intergrain phase
                state.material['Ms'] = Ms_values.take(voi.domains)

        *Arguments*
            mesh ([:class:`Mesh`])
                Mesh object which should be tesselated
            num_grains (int)
                Number of grains to be created
            seed_points ([:class:`torch.Tensor`])
                User provided seed points of size (num_grains, 3)
        """
        if num_grains != None and seed_points == None:
            L = torch.tensor(mesh.dx)*torch.tensor(mesh.n)
            offset = torch.tensor(mesh.origin)
            self._points = L * torch.rand((num_grains, 3)) + offset
        elif num_grains == None and seed_points != None:
            self._points = seed_points
            num_grains = int(seed_points.shape[0])
        else:
            raise ValueError("Either 'num_grains' or 'seed_points' need to be specified!")

        self._grid = torch.stack(mesh.SpatialCoordinate(),dim=-1).reshape(-1, 3)
        self._mesh = mesh

        self._points = self._points.to(dtype=torch.float32)
        self._grid = self._grid.to(dtype=torch.float32)
        self._update_domains()

        logging.info_green("[Voronoi] Setup initial Tesselation (num_grains = %d)" % (num_grains))

    @property
    def domains(self):
        return self._domains.reshape(self._mesh.n)

    @property
    def points(self):
        return self._points

    def _update_points(self):
        ## orginal non-vectorized code (factor 2 slower)
        #for i in range(self._points.shape[0]):
        #    mask = self._domains == i
        #    if mask.sum() > 0:
        #        centroid = self._grid[mask].mean(dim=0)
        #        self._points[i] = centroid

        new_points = torch.zeros_like(self._points)
        counts = torch.zeros(self.points.shape[0])

        # Scatter add the grid points to their corresponding centroid accumulators
        new_points.index_add_(0, self._domains, self._grid)

        # Count the number of points in each Voronoi cell
        ones = torch.ones(self._grid.size(0))
        counts.index_add_(0, self._domains, ones)

        # Avoid division by zero
        valid_mask = counts > 0
        new_points[valid_mask] /= counts[valid_mask].unsqueeze(1)

        self._points = new_points

    def _update_domains(self):
        ## original vectorized code (requries N_mesh * N_points memory)
        #distances = torch.cdist(self._grid, self._points)
        #self._domains = distances.argmin(dim=1)

        tree = KDTree(self._points.cpu().numpy())
        dd, ii = tree.query(self._grid.cpu().numpy())
        self._domains = torch.tensor(ii)

    def relax(self, it = 5):
        for i in range(it):
            self._update_points()
            self._update_domains()
            logging.info_blue("[Voronoi] Relax Tesselation (Lloyd's Iteration %d)" % i)
        return self.domains

    def add_intergrain_phase(self, thickness):
        domains = self.domains
        intergrain_id = domains.max() + 1
        grad = domains
        dim = [i for i in range(3) if domains.shape[i] > 1]

        for i in range(thickness):
            grad = torch.gradient(grad, dim=dim)
            grad = sum([g.abs() for g in grad])
            domains[grad > 1e-15] = intergrain_id



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import torch
from magnumnp.common import logging

__all__ = ["Material"]


class Material(dict):
    def __init__(self, state):
        self._state = state

    def __getitem__(self, key):
        return super().__getitem__(key)(self._state)

    def __setitem__(self, key, value):
        if callable(value):
            if not isinstance(value(self._state), torch.Tensor):
                logging.warning("Casting of state-dependent material parameters requires conversion in every call.")
                logging.warning("Use state.Constant(value) inside of your function in order to prevent this.")
            super().__setitem__(key, lambda state: self._state.convert_tensorfield(value(state)))
        else:
            t = self._state.convert_tensorfield(value)
            super().__setitem__(key, lambda state: t) # allow constant material parameters to be called

    def set(self, material, domain=None):
        r"""
        Setting several constant material parameters at once

        :param materials: material that should be set
        :type materials:  :class:`Material`

        :param domains: domain where the materials should be set
        :type domains:  :class:`torch.Tensor`, optional

        :Example:

            .. code::

            # set material everywhere
            state.material.set(material0)

            # set material in certain domain
            state.material.set(material1, domain1)
        """
        for key, value in material.items():
            if domain == None:
                self[key] = self._state.Constant(value)
            else:
                if key not in self.keys():
                    self[key] = self._state.Constant(value)
                    self[key][...] = 0.
                self[key][domain] = value



from magnumnp.common import logging, timedmethod
import torch
import numpy as np

__all__ = ["Minimizer_LBFGS"]

class Minimizer_LBFGS(object):
    def __init__(self, terms, tolerance= 1e-8, max_iter=1000, memory_size = 5, dm_max = 1e4):
        self._terms = terms
        self._dm_max = dm_max
        self._memory_size = memory_size
        self._tolerance = tolerance
        self.max_iter = max_iter

    def _dm(self, state):
        h = sum([term.h(state) for term in self._terms])
        return torch.linalg.cross(state.m, torch.linalg.cross(state.m, h))
    
    def linesearch(self, state, p):
        alpha = 1.0 #initial step size
        tau = 0.5  # Reduction factor
        c = 0.1  # Sufficient decrease parameter   
        h = sum([term.h(state) for term in self._terms])
        dm = torch.linalg.cross(state.m, torch.linalg.cross(state.m, h))
        E = sum([term.E(state) for term in self._terms])
        m = (dm*p).sum()
        t = -c*m
        j = 0
        maxIter = 1000
        
        while j < maxIter:
            m_new = state.m + alpha * p
            state.m = m_new
            E_new = sum([term.E(state) for term in self._terms])
            sufficient_decrease = E - E_new - alpha * t
            
            if sufficient_decrease >= 0:
                break
            
            alpha *= tau
            j += 1
    
        return state

    def minimize(self, state, tolerance= 1e-8, max_iter=1000, memory_size = 5,):
        step = 0
        glob_step = 0
        glob_steps = []
        glob_steps.append(glob_step)
        E = sum([term.E(state) for term in self._terms]) #Energy
        energy = []
        energy.append(E)
        h = sum([term.h(state) for term in self._terms])
        s_vectors = [torch.zeros(state.m.size())]*memory_size
        y_vectors = [torch.zeros(state.m.size())]*memory_size
        dm = torch.linalg.cross(state.m, torch.linalg.cross(state.m, h)) #Gradient

        eps = 2.22e-16;
        eps2 = np.sqrt(eps)
        epsr = np.power(eps, 0.9)
        tolf2 = np.sqrt(tolerance)
        tolf3 = np.power(tolerance, 0.3333333333333333333333333)

        dm_max = dm.max()
        if dm_max < epsr*(1+torch.abs(E)):
            return state, energy, glob_steps
        
        alpha = np.zeros(memory_size, dtype=np.float64)
        q = torch.zeros(state.m.size(), dtype=torch.float64)
        s = torch.zeros(state.m.size(), dtype=torch.float64)
        y = torch.zeros(state.m.size(), dtype=torch.float64)
        E_old = 0.0
        dm_old = torch.zeros(dm.size(), dtype=torch.float64)
        m_old = state.m;
        H0k = 1

        while glob_step < max_iter:
            
            E_old = E
            m_old = state.m
            dm_old = dm
            
            q = dm.clone()
            k = min(memory_size, step)
            for i in range(k - 1, -1, -1):
                rho = 1.0 / (s_vectors[i]* y_vectors[i]).sum()
                alpha[i] = rho * (s_vectors[i]*q).sum()
                q -= alpha[i] * y_vectors[i]
                
            z = H0k * q
            for i in range(k):
                rho = 1.0 / (s_vectors[i]* y_vectors[i]).sum()
                beta = rho * (y_vectors[i]*z).sum()
                z += s_vectors[i] * (alpha[i] - beta)
                #z = -z for minimization?
                
            phiPrime0 = -(dm*z).sum()
            if phiPrime0 > 0:
                z = dm
                step = 0
 
            #linesearch and update of state.m
            state = self.linesearch(state, z)
            
            #update gradient
            h = sum([term.h(state) for term in self._terms])
            dm = torch.linalg.cross(state.m, torch.linalg.cross(state.m, h))

            dm_max = dm.max()
            if dm_max < epsr*(1+torch.abs(E)):
                return state, energy, glob_steps
            
            m_diff = state.m - m_old
            dm_diff = dm - dm_old
            
            ys = (dm_diff*m_diff).sum()
            #if ys <= eps2 * torch.norm(dm_diff) * torch.norm(m_diff):
            #    
            #else:
            if step < memory_size:
                s_vectors[step] = m_diff
                y_vectors[step] = dm_diff
            else:
                s_vectors[:-1] = s_vectors[1:]  # Cyclic shift of s_vectors
                s_vectors[memory_size - 1] = m_diff
                y_vectors[:-1] = y_vectors[1:]  # Cyclic shift of y_vectors
                y_vectors[memory_size - 1] = dm_diff
            H0k = ys / (dm_diff*dm_diff).sum()
            step += 1
            
            # increase step count
            glob_step += 1
            glob_steps.append(glob_step)
            
        return state, energy, glob_steps





#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from magnumnp.common import logging, timedmethod, constants, normalize
from .ode_solvers import RKF45
import torch

__all__ = ["LLGSolver"]

class LLGSolver(object):
    def __init__(self, terms, solver = RKF45, no_precession = False, **kwargs):
        """
        This class implements the LLG dm term as well as the corresponding energy. 
        It also provides the interface for time-integration and allows switching between
        different ODE solvers.

        *Example*
            .. code:: python

            llg = LLGSolver([demag, exchange, external])
            logger = Logger("data", ['t', 'm'])
            while state.t < 1e-9-eps:
                llg.step(state, 1e-11)
                logger << state

        *Arguments*
            terms ([:class:`LLGTerm`])
                List of LLG contributions to be considered for time integration
            solver ([:class:`Solver`])
                ODE solver to be used (chose one of RKF45 (default), RKF56, ScipyODE, ScipyOdeint, TorchDiffEq, TorchDiffEqAdjoint)
            no_precession (bool)
                integrate without precession term (default: False)
        """
        self._terms = terms
        self._solver = solver(self.dm, **kwargs)
        self._no_precession = no_precession

    def dm(self, t, x, state, alpha = None):
        state.t = t
        state.m = x
        alpha = alpha or state.material["alpha"]

        gamma_prime = constants.gamma / (1. + alpha**2)
        alpha_prime = alpha * gamma_prime

        h = sum([term.h(state) for term in self._terms])

        cross_m_h = torch.linalg.cross(state.m, h)

        dm = -alpha_prime * torch.linalg.cross(state.m, cross_m_h)
        if not self._no_precession:
            dm -= gamma_prime * cross_m_h

        return dm

    def E(self, state):
        return sum([term.E(state) for term in self._terms])

    @timedmethod
    def step(self, state, dt, **kwargs):
        state.t, state.m = self._solver.step(state.t, state.m, dt, state=state, **kwargs)
        normalize(state.m)
        logging.info_blue("[LLG] step: dt= %g  t=%g" % (dt, state.t))

    @timedmethod
    def solve(self, state, tt, **kwargs):
        logging.info_blue("[LLG] solve: t0=%g  t1=%g Integrating ..." % (tt[0].cpu().numpy(), tt[-1].cpu().numpy()))
        res = self._solver.solve(tt, state.m, state=state, **kwargs)
        logging.info_blue("[LLG] solve: t0=%g  t1=%g Finished" % (tt[0].cpu().numpy(), tt[-1].cpu().numpy()))

        state.t = tt[-1]
        state.m = res[-1]
        return res

    @timedmethod
    def relax(self, state, maxiter = 500, dm_tol = 1e2, dt = 1e-11):
        t0 = state.t

        for i in range(maxiter):
            state.t, state.m = self._solver.step(state.t, state.m, dt, state=state, alpha = 1.0) #, no_precession = True) # no_precession requires more iterations for SP4 demo!?

            dm = self.dm(state.t, state.m, state=state, alpha = 1.0).abs().max() / constants.gamma # use same scaling as within minimizer
            logging.info_blue("[LLG] relax: i=%d t=%g |dm|=%g" % (i, state.t-t0, dm))
            if dm < dm_tol:
                logging.info_green("[LLG] relax: Successfully converged (iter=%d, dm_tol = %g)" % (i, dm_tol))
                state.t = t0
                return True

        logging.warning("[LLG] relax: Terminated after maxiter = %d (dm = %g, dm_tol = %g)" % (maxiter, dm, dm_tol))
        state.t = t0
        return False



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from magnumnp.common import logging, timedmethod, constants, Timer, complex_dtype
from .field_terms import LinearFieldTerm
import numpy as np
import torch
import torch.fft
from torch import asinh, atan, sqrt, log, abs, pi
from time import time
import os

__all__ = ["DemagField"]

def f(x, y, z):
    x, y, z = abs(x), abs(y), abs(z)
    x2, y2, z2 = x**2, y**2, z**2
    r = sqrt(x2 + y2 + z2)
    res = 1.0 / 6.0 * (2*x2 - y2 - z2) * r
    res += (y / 2.0 * (z2 - x2) * asinh(y / sqrt(x2 + z2))).nan_to_num(posinf=0, neginf=0)
    res += (z / 2.0 * (y2 - x2) * asinh(z / sqrt(x2 + y2))).nan_to_num(posinf=0, neginf=0)
    res -= (x * y * z * atan(y*z / (x * r))).nan_to_num(posinf=0, neginf=0)
    return res

def g(x, y, z):
    z = abs(z)
    x2, y2, z2 = x**2, y**2, z**2
    r = sqrt(x2 + y2 + z2)
    res = -x * y * r / 3.0
    res += (x * y * z * asinh(z / sqrt(x2 + y2))).nan_to_num(posinf=0, neginf=0)
    res += (y / 6.0 * (3.0 * z2 - y2) * asinh(x / sqrt(y2 + z2))).nan_to_num(posinf=0, neginf=0)
    res += (x / 6.0 * (3.0 * z2 - x2) * asinh(y / sqrt(x2 + z2))).nan_to_num(posinf=0, neginf=0)
    res -= (z**3 / 6.0 * atan(x * y / (z * r))).nan_to_num(posinf=0, neginf=0)
    res -= (z * y2 / 2.0 * atan(x * z / (y * r))).nan_to_num(posinf=0, neginf=0)
    res -= (z * x2 / 2.0 * atan(y * z / (x * r))).nan_to_num(posinf=0, neginf=0)
    return res

def F1(func, x, y, z, dz, dZ):
    return func(x, y, z      + dZ) \
         - func(x, y, z          ) \
         - func(x, y, z - dz + dZ) \
         + func(x, y, z - dz     )

def F0(func, x, y, z, dy, dY, dz, dZ):
    return F1(func, x, y      + dY, z, dz, dZ) \
         - F1(func, x, y,           z, dz, dZ) \
         - F1(func, x, y - dy + dY, z, dz, dZ) \
         + F1(func, x, y - dy,      z, dz, dZ)

def newell(func, x, y, z, dx, dy, dz, dX, dY, dZ):
    res = F0(func, x,           y, z, dy, dY, dz, dZ) \
        - F0(func, x - dx,      y, z, dy, dY, dz, dZ) \
        - F0(func, x + dX,      y, z, dy, dY, dz, dZ) \
        + F0(func, x - dx + dX, y, z, dy, dY, dz, dZ)
    return -res / (4.*pi*dx*dy*dz)

def dipole_f(x, y, z, dx, dy, dz, dX, dY, dZ):
    z = z + dZ/2. - dz/2. # diff of cell centers for non-equidistant demag
    res = (2.*x**2 - y**2 - z**2) * pow(x**2 + y**2 + z**2, -5./2.)
    res[0,0,0] = 0.
    return res * dx*dy*dz / (4.*pi)

def dipole_g(x, y, z, dx, dy, dz, dX, dY, dZ):
    z = z + dZ/2. - dz/2. # diff of cell centers for non-equidistant demag
    res = 3.*x*y * pow(x**2 + y**2 + z**2, -5./2.)
    res[0,0,0] = 0.
    return res * dx*dy*dz / (4.*pi)

def demag_f(x, y, z, dx, dy, dz, dX, dY, dZ, p):
    res = dipole_f(x, y, z, dx, dy, dz, dX, dY, dZ)
    near = (x**2 + y**2 + z**2) / max(dx**2 + dy**2 + dz**2, dX**2 + dY**2 + dZ**2) < p**2
    res[near] = newell(f, x[near], y[near], z[near], dx, dy, dz, dX, dY, dZ)
    return res

def demag_g(x, y, z, dx, dy, dz, dX, dY, dZ, p):
    res = dipole_g(x, y, z, dx, dy, dz, dX, dY, dZ)
    near = (x**2 + y**2 + z**2) / max(dx**2 + dy**2 + dz**2, dX**2 + dY**2 + dZ**2) < p**2
    res[near] = newell(g, x[near], y[near], z[near], dx, dy, dz, dX, dY, dZ)
    return res


class DemagField(LinearFieldTerm):
    r"""
    Demagnetization Field:

    The dipole-dipole interaction gives rise to a long-range interaction.
    The integral formulation of the corresponding Maxwell equations can
    be represented as convolution of the magnetization :math:`\vec{M} = M_s \; \vec{m}` with a proper
    demagnetization kernel :math:`\vec{N}`

    .. math::
        \vec{h}^\text{dem}_{\vec{i}} = \sum\limits_{\vec{j}} \vec{N}_{\vec{i} - \vec{j}} \, \vec{M}_{\vec{j}},

    The convolution can be evaluated efficiently using an FFT method.

    :param p: number of next neighbors for near field via Newell's equation (default = 20)
    :type p: int, optional
    """
    def __init__(self, p = 20, cache_dir = None):
        self._p = p
        self._cache_dir = cache_dir

    def _shape(self, state): # TODO: try padding to 2N-1 for small N like mumax does
        s = [1,1,1]
        for i in range(3):
            if state.mesh.n[i] == 1:
                continue
            if state.mesh.pbc[i] == 0:
                s[i] = 2*state.mesh.n[i]
            else:
                s[i] = state.mesh.n[i] # no need to pad if nonzero pbc
        return s

    def _init_N_component(self, state, perm, func):
        dx = np.array(state.mesh.dx)
        dx /= dx.min() # rescale dx to avoid NaNs when using single precision

        shape = self._shape(state)
        ij = [torch.fft.fftfreq(n,1/n) for n in shape] # local indices
        ij = torch.meshgrid(*ij,indexing='ij')
        x, y, z = [ij[ind]*dx[ind] for ind in perm]
        Lx = [state.mesh.n[ind]*dx[ind] for ind in perm]
        dx = [dx[ind] for ind in perm]

        offsets = [torch.arange(-state.mesh.pbc[ind], state.mesh.pbc[ind]+1) for ind in perm] # offset of pseudo PBC images
        offsets = torch.stack(torch.meshgrid(*offsets, indexing="ij"), dim=-1).flatten(end_dim=-2)

        Nc = torch.zeros(shape)
        for offset in offsets:
            Nc += func(x + offset[0]*Lx[0], y + offset[1]*Lx[1], z + offset[2]*Lx[2], *dx, *dx, self._p)

        dim = [i for i in range(3) if state.mesh.n[i] > 1]
        if len(dim) > 0:
            Nc = torch.fft.rfftn(Nc, dim = dim)
        return Nc.real.clone()


    def _init_N(self, state):
        name = "/N_%s.pt" % str(state.mesh).replace(" ","")
        if self._cache_dir != None and os.path.isfile(self._cache_dir + name):
            [Nxx,Nxy,Nxz,Nyy,Nyz,Nzz] = torch.load(self._cache_dir + name, map_location=state.device)
            logging.info("[DEMAG]: Use cached demag kernel from '%s'" % (self._cache_dir + name))
        else:
            dtype = torch.get_default_dtype()
            torch.set_default_dtype(torch.float64) # always use double precision
            time_kernel = time()

            Nxx = self._init_N_component(state, [0,1,2], demag_f).to(dtype=dtype)
            Nxy = self._init_N_component(state, [0,1,2], demag_g).to(dtype=dtype)
            Nxz = self._init_N_component(state, [0,2,1], demag_g).to(dtype=dtype)
            Nyy = self._init_N_component(state, [1,2,0], demag_f).to(dtype=dtype)
            Nyz = self._init_N_component(state, [1,2,0], demag_g).to(dtype=dtype)
            Nzz = self._init_N_component(state, [2,0,1], demag_f).to(dtype=dtype)

            logging.info(f"[DEMAG]: Time calculation of demag kernel = {time() - time_kernel} s")
            torch.set_default_dtype(dtype) # restore dtype

            # cache demag tensor
            if self._cache_dir != None:
                if not os.path.isdir(self._cache_dir):
                    os.makedirs(self._cache_dir)
                torch.save([Nxx,Nxy,Nxz,Nyy,Nyz,Nzz], self._cache_dir + name)
                logging.info("[DEMAG]: Save demag kernel to '%s'" % (self._cache_dir + name))

        return [[Nxx, Nxy, Nxz],
                [Nxy, Nyy, Nyz],
                [Nxz, Nyz, Nzz]]


    @timedmethod
    def h(self, state):
        if not hasattr(self, "_N"):
            self._N = self._init_N(state)

        dim = [i for i in range(3) if state.mesh.n[i] > 1]
        shape = self._shape(state)
        s = [shape[i] for i in dim]

        if len(dim) == 0: # single spin   TODO: remove this when torch issue #96518 has been solved
            N = torch.stack([torch.stack(self._N[0], dim=-1),
                             torch.stack(self._N[1], dim=-1),
                             torch.stack(self._N[2], dim=-1)], dim=-1)
            return (N * state.m).sum(dim=-1)

        hx = torch.zeros(self._N[0][0].shape, dtype=complex_dtype[self._N[0][0].dtype], device=state.device)
        hy = torch.zeros(self._N[0][0].shape, dtype=complex_dtype[self._N[0][0].dtype], device=state.device)
        hz = torch.zeros(self._N[0][0].shape, dtype=complex_dtype[self._N[0][0].dtype], device=state.device)

        for ax in range(3):
            m_pad_fft1D = torch.fft.rfftn(state.material["Ms"] * state.m[:,:,:,(ax,)], dim = dim, s = s).squeeze(-1)

            hx += self._N[0][ax] * m_pad_fft1D
            hy += self._N[1][ax] * m_pad_fft1D
            hz += self._N[2][ax] * m_pad_fft1D

        hx = torch.fft.irfftn(hx, dim = dim)
        hy = torch.fft.irfftn(hy, dim = dim)
        hz = torch.fft.irfftn(hz, dim = dim)

        return torch.stack([hx[:state.mesh.n[0],:state.mesh.n[1],:state.mesh.n[2]],
                            hy[:state.mesh.n[0],:state.mesh.n[1],:state.mesh.n[2]],
                            hz[:state.mesh.n[0],:state.mesh.n[1],:state.mesh.n[2]]], dim=3)

    def E(self, state, domain = Ellipsis): # TODO: remove as soon as @compile works for DemagField
        E = -0.5 * constants.mu_0 * state.material["Ms"] * state.m * self.h(state) * state.mesh.cell_volumes
        return E[domain].sum()



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import numpy as np
import torch
from scipy import interpolate

__all__ = ["TimeInterpolator"]

class TimeInterpolator(object):
    def __init__(self, state, points, order = 1):
        self._tp = torch.tensor(list(points.keys()))
        self._fp = [state.convert_tensorfield(f) for f in points.values()]
        self._state = state
        self._order = order

    def __call__(self, state):
        i = torch.searchsorted(self._tp, state.t) # upper index
        tp = self._tp
        fp = self._fp
        if self._order == 0:
            i = torch.clamp(i, min=1, max=len(self._tp)) # extrapolate only on lower bound
            return fp[i-1]
        elif self._order == 1:
            i = torch.clamp(i, min=1, max=len(self._tp)-1) # extrapolate on bounds
            return fp[i-1] + (state.t-tp[i-1]) / (tp[i]-tp[i-1]) * (fp[i] - fp[i-1])
        else:
            raise ValueError("Order must be 0 or 1(default)")

    @property
    def final_time(self):
        return self._tp[-1]



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

mu_0 = 1.2566370614e-6
"""magnetic constant mu_0"""

gamma = 2.21276157e5
"""gyromagnetic ratio gamma"""

mu_B = 9.2740154e-24
"""Bohr magneton mu_bohr"""

e = 1.602176487e-19
"""elementary charge e"""

kb = 1.380648813e-23
"""Boltzmann constant kb"""

hbar = 1.0545718e-34
"""reduced Planck constant"""

me = 9.1093837e-31
"""Electron Mass"""

c = 299792458
"""Speed of Light"""



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from magnumnp.common import timedmethod, constants
import torch
from .field_terms import LinearFieldTerm
import numpy as np

__all__ = ["DMIField", "InterfaceDMIField", "BulkDMIField", "D2dDMIField"]


class DMIField(LinearFieldTerm):
    r"""
    General Dzyaloshinskii-Moriya interaction

    The general expression for the DMI field can be expressed as

    .. math::
        \vec{h}^\text{dmi}(\vec{x}) = \frac{2 \, D}{\mu_0 \, M_s} \; \sum_{k=x,y,z} \vec{e}^\text{dmi}_k \times \frac{\partial \vec{m}}{\partial \vec{e}_k},

    with the DMI strength :math:`D` and the DMI vectors :math:`\vec{e}^\text{dmi}_k`, which describe which components of the gradient of :math:`\vec{m}` contribute to which component of the corresponding field. It is assumed that :math:`\vec{e}^\text{dmi}_{-k} = -\vec{e}^\text{dmi}_k`.

    The occuring gradient is discretized using central differences which finally yields

    .. math::
        \vec{h}^\text{dmi}_i = \frac{2 \, D_i}{\mu_0 \, M_{s,i}} \; \sum_{k=\pm x, \pm y,\pm z} \frac{\vec{e}^\text{dmi}_k \times \vec{m}_{i+\vec{e}_k}}{2 \, \Delta_k}.

    :param Ku: Name of the material parameter for the anisotropy constant :math:`K_\text{u}`, defaults to "Ku"
    :type Ku: str, optional
    :param Ku_axis: Name of the material parameter for the anisotropy axis :math:`\vec{e}_\text{u}`, defaults to "Ku_axis"
    :type Ku_axis: str, optional
    """
    parameters = ["D"]
    def __init__(self, dmi_vector, **kwargs):
        self._dmi_vector = torch.tensor(dmi_vector)
        super().__init__(**kwargs)

    @timedmethod
    @torch.compile
    def h(self, state):
        D = state.material[self.D]
        Ms = state.material["Ms"]
        m = state.m
        dx = state.mesh.dx_tensor[0].reshape(-1,1,1,1)
        dy = state.mesh.dx_tensor[1].reshape(1,-1,1,1)
        dz = state.mesh.dx_tensor[2].reshape(1,1,-1,1)
        h = torch.zeros_like(state.m)

        # x
        if state.mesh.pbc[0] == 0:
            v = self._dmi_vector[0].expand(m[1:,:,:].shape)
            # TODO: implement for different signs (according to https://github.com/mumax/3/issues/236)
            # D_avg = torch.where(D[1:,:,:]*D[:-1,:,:] < 0, ...
            D_avg = 2.*D[1:,:,:]*D[:-1,:,:] / (D[1:,:,:]*dx[:-1,:,:,:] + D[:-1,:,:]*dx[1:,:,:,:])
            h[:-1,:,:] += D_avg * torch.linalg.cross(v, m[ 1:,:,:]) / 2.
            h[ 1:,:,:] -= D_avg * torch.linalg.cross(v, m[:-1,:,:]) / 2.
        else:
            v = self._dmi_vector[0].expand(m.shape)
            D_next = torch.roll(D, +1, dims=0)
            dx_next = torch.roll(dx, +1, dims=0)
            # TODO: implement for different signs (according to https://github.com/mumax/3/issues/236)
            # D_avg = torch.where(D_next*D < 0,
            D_avg = 2.*D_next*D / (D_next*dx + D*dx_next)
            h -= D_avg * torch.linalg.cross(v, torch.roll(state.m, +1, dims=0)) / 2.

            D_avg = torch.roll(D_avg, -1, dims=0)
            h += D_avg * torch.linalg.cross(v, torch.roll(state.m, -1, dims=0)) / 2.

        # y
        if state.mesh.pbc[1] == 0:
            v = self._dmi_vector[1].expand(m[:,1:,:].shape)
            # TODO: implement for different signs (according to https://github.com/mumax/3/issues/236)
            # D_avg = torch.where(D[:,1:,:]*D[:,:-1,:] < 0,
            D_avg = 2.*D[:,1:,:]*D[:,:-1,:] / (D[:,1:,:]*dy[:,:-1,:,:] + D[:,:-1,:]*dy[:,1:,:,:])
            h[:,:-1,:] += D_avg * torch.linalg.cross(v, m[:, 1:,:]) / 2.
            h[:, 1:,:] -= D_avg * torch.linalg.cross(v, m[:,:-1,:]) / 2.
        else:
            v = self._dmi_vector[1].expand(m.shape)
            D_next = torch.roll(D, +1, dims=1)
            dy_next = torch.roll(dx, +1, dims=1)
            # TODO: implement for different signs (according to https://github.com/mumax/3/issues/236)
            # D_avg = torch.where(D_next*D < 0,
            D_avg = 2.*D_next*D / (D_next*dy + D*dy_next)
            h -= D_avg * torch.linalg.cross(v, torch.roll(state.m, +1, dims=1)) / 2.

            D_avg = torch.roll(D_avg, -1, dims=1)
            h += D_avg * torch.linalg.cross(v, torch.roll(state.m, -1, dims=1)) / 2.

        # z
        if state.mesh.pbc[2] == 0:
            v = self._dmi_vector[2].expand(m[:,:,1:].shape)
            # TODO: implement for different signs (according to https://github.com/mumax/3/issues/236)
            # D_avg = torch.where(D[:,:,1:]*D[:,:,:-1] < 0,
            D_avg = 2.*D[:,:,1:]*D[:,:,:-1] / (D[:,:,1:]*dz[:,:,:-1,:] + D[:,:,:-1]*dz[:,:,1:,:])
            h[:,:,:-1] += D_avg * torch.linalg.cross(v, m[:,:, 1:]) / 2.
            h[:,:, 1:] -= D_avg * torch.linalg.cross(v, m[:,:,:-1]) / 2.
        else:
            v = self._dmi_vector[2].expand(m.shape)
            D_next = torch.roll(D, +1, dims=2)
            dz_next = torch.roll(dx, +1, dims=2)
            # TODO: implement for different signs (according to https://github.com/mumax/3/issues/236)
            # D_avg = torch.where(D_next*D < 0,
            D_avg = 2.*D_next*D / (D_next*dz + D*dz_next)
            h -= D_avg * torch.linalg.cross(v, torch.roll(state.m, +1, dims=2)) / 2.

            D_avg = torch.roll(D_avg, -1, dims=2)
            h += D_avg * torch.linalg.cross(v, torch.roll(state.m, -1, dims=2)) / 2.

        h *= 2. / (constants.mu_0 * Ms)
        return h.nan_to_num(posinf=0, neginf=0)



class InterfaceDMIField(DMIField):
    r"""
    Interface Dzyaloshinskii-Moriya interaction

    .. math::
        \vec{h}^\text{dmii}(\vec{x}) = -\frac{2 \, D_i}{\mu_0 \, M_s} \; \left[ \nabla \left(\vec{e}_z \cdot \vec{m} \right) - \left(\nabla \cdot \vec{m} \right) \, \vec{e}_z\right],

    with the DMI strength :math:`D_i` and an interface normal :math:`\vec{e}_z` in z-direction.
    The corresponding DMI vectors are :math:`\vec{e}^\text{dmi}_x = [ 0, 1, 0]`, :math:`\vec{e}^\text{dmi}_y = [-1, 0, 0]`, and :math:`\vec{e}^\text{dmi}_z = [0, 0, 0]`.

    :param Di: Name of the material parameter for the anisotropy constant :math:`D_i`, defaults to "Di"
    :type Di: str, optional
    """
    def __init__(self, Di = "Di"):
        dmi_vector = [[ 0., 1., 0.], # x
                      [-1., 0., 0.], # y
                      [ 0., 0., 0.]] # z
        super().__init__(dmi_vector, D = Di)


class BulkDMIField(DMIField):
    r"""
    Bulk Dzyaloshinskii-Moriya interaction

    .. math::
        \vec{h}^\text{dmib}(\vec{x}) = -\frac{2 \, D_b}{\mu_0 \, M_s} \; \nabla \times \vec{m},

    with the DMI strength :math:`D_b`.
    The corresponding DMI vectors are :math:`\vec{e}^\text{dmi}_x = [1, 0, 0]`, :math:`\vec{e}^\text{dmi}_y = [0, 1, 0]`, and :math:`\vec{e}^\text{dmi}_z = [0, 0, 1]`.

    :param Db: Name of the material parameter for the anisotropy constant :math:`D_i`, defaults to "Di"
    :type Db: str, optional
    """
    def __init__(self, Db = "Db"):
        dmi_vector = [[1., 0., 0.], # x
                      [0., 1., 0.], # y
                      [0., 0., 1.]] # z
        super().__init__(dmi_vector, D = Db)


class D2dDMIField(DMIField):
    r"""
    D2d Dzyaloshinskii-Moriya interaction

    .. math::
        \vec{h}^\text{dmib}(\vec{x}) = -\frac{2 \, D_b}{\mu_0 \, M_s} \; \nabla \times \vec{m},

    with the DMI strength :math:`D_{D2d}`.
    The corresponding DMI vectors are :math:`\vec{e}^\text{dmi}_x = [-1, 0, 0]`, :math:`\vec{e}^\text{dmi}_y = [0, 1, 0]`, and :math:`\vec{e}^\text{dmi}_z = [0, 0, 0]`.

    :param DD2d: Name of the material parameter for the anisotropy constant :math:`D_{D2d}`, defaults to "DD2d"
    :type DD2d: str, optional
    """
    def __init__(self, DD2d = "DD2d"):
        dmi_vector = [[-1., 0., 0.], # x
                      [ 0., 1., 0.], # y
                      [ 0., 0., 0.]] # z
        super().__init__(dmi_vector, D = DD2d)



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import torch
import numpy as np
import scipy
import pyvista as pv
import os
from . import Mesh
from magnumnp.common import logging, Material

__all__ = ["write_vtr", "write_vti", "read_vti", "read_image", "read_mesh", "read_neper"]

def write_vtr(fields, filename, state = None, scale = 1.):
    if not filename.endswith(".vtr"):
        logging.warning("[write_vtr] Extention '.vtr' should be used on non-equidistant grids! (filename = '%s')" % filename)

    dirname = os.path.dirname(filename)
    if dirname and not os.path.isdir(dirname):
        os.makedirs(dirname)

    if not (isinstance(fields, list) or isinstance(fields, dict)):
        fields = [fields]
    if not isinstance(fields, dict):
        fields = {"f%03d"%i:f for (i,f) in enumerate(fields)}

    if state is None:
        n = list(fields.values())[0].shape[:3]
        dx = (1., 1., 1.)
        origin = (0., 0., 0.)
    else:
        n = state.mesh.n
        dx = state.mesh.dx
        origin = state.mesh.origin

    x = torch.hstack([torch.tensor([0.]), state.mesh.dx_tensor[0].cumsum(0)]).cpu().numpy() + state.mesh.origin[0]
    y = torch.hstack([torch.tensor([0.]), state.mesh.dx_tensor[1].cumsum(0)]).cpu().numpy() + state.mesh.origin[1]
    z = torch.hstack([torch.tensor([0.]), state.mesh.dx_tensor[2].cumsum(0)]).cpu().numpy() + state.mesh.origin[2]

    grid = pv.RectilinearGrid(x*scale, y*scale, z*scale)

    for name in fields:
        f = fields[name]
        if len(f.shape) == 0 or len(f.shape) == 1: # expand constant tensor to tensorfield
            f = f.expand(n + f.shape)
        if len(f.shape) == 4 and f.shape[-1] == 1: # remove dim for scalar field (nx,ny,nz,1) => (nx,ny,nz)
            f = f[: ,:, :, 0]
        if len(f.shape) == 3: # scalar data
            grid.cell_data.set_array(f.detach().cpu().numpy().flatten('F'), name)
        elif len(f.shape) == 4: # vector data
            grid.cell_data.set_array(f.detach().cpu().numpy().reshape(-1,3,order='F'), name)
        else:
            raise ValueError("write_vti: unsupported data format (", name, f.shape, ")")
    grid.save(filename)


def write_vti(fields, filename, state = None, scale = 1.):
    r"""
    Write vti files (equidistant rectangular grid, compressed) using pyvista.

    :param fields: single torch Tensor or List/Dictionary of tensors to be written
    :type fields: :class:`Tensor`, list, dict
    :param filename: filename to be written
    :type filename: str
    :param state: filename to be written
    :type state: :class:`State`

    :Examples:

    .. code::

        # write single scalar or vector
        write_vti(state.material.Ms, "scalar.vti")
        write_vti(state.m, "vector.vti")

        # use dictinary or list
        write_vti([state.m, h], "list.vti")
        write_vti({'m':state.m, 'h':h}, "dict.vti")
    """
    if str(filename)[-4:] != ".vti":
        logging.warning("[write_vti] Extention '.vti' should be used on equidistant grids!")

    dirname = os.path.dirname(filename)
    if dirname and not os.path.isdir(dirname):
        os.makedirs(dirname)

    if not (isinstance(fields, list) or isinstance(fields, dict)):
        fields = [fields]
    if not isinstance(fields, dict):
        fields = {"f%03d"%i:f for (i,f) in enumerate(fields)}

    if state is None:
        values = [fields[name] for name in fields]
        n = values[0].shape[:3]
        dx = (1., 1., 1.)
        origin = (0., 0., 0.)
    else:
        n = state.mesh.n
        dx = state.mesh.dx
        origin = state.mesh.origin

    grid = pv.ImageData(dimensions = np.array(n) + 1,
                        spacing = np.array(dx) * scale,
                        origin = np.array(origin) * scale)

    for name in fields:
        f = fields[name]
        if len(f.shape) == 0 or len(f.shape) == 1: # expand constant tensor to tensorfield
            f = f.expand(n + f.shape)
        if len(f.shape) == 4 and f.shape[-1] == 1: # remove dim for scalar field (nx,ny,nz,1) => (nx,ny,nz)
            f = f[: ,:, :, 0]
        if len(f.shape) == 3: # scalar data
            grid.cell_data.set_array(f.detach().cpu().numpy().flatten('F'), name)
        elif len(f.shape) == 4: # vector data
            grid.cell_data.set_array(f.detach().cpu().numpy().reshape(-1,3,order='F'), name)
        else:
            raise ValueError("write_vti: unsupported data format (", name, f.shape, ")")

    grid.save(filename)


def read_vti(filename, scale = 1.):
    r"""
    Read vti files using pyvista

    :param str filename: Filename to be read
    :param float scale: scale with which the file was written
    :return :class:`Mesh` & dict: Mesh object and dictionary containing all data tensors


    :Examples:
      .. code::
        mesh, fields = read_vti("m0.vti")
    """
    fields = {}
    data = pv.read(filename)

    mesh = Mesh(np.array(data.dimensions)-1, np.array(data.spacing) / scale, np.array(data.origin) / scale)

    for name in data.array_names:
        f = data.get_array(name)
        vals = data.get_array(name)
        if len(vals.shape) == 1:
            dim = mesh.n
        else:
            dim = mesh.n + (vals.shape[-1],)
        t = torch.tensor(1.)
        f = torch.from_numpy(vals.reshape(dim, order="F")).to(device=t.device, dtype=t.dtype)
        fields[name] = f
    return mesh, fields


#TODO: move to utils since it depends on numpy + scipy
def read_image(mesh, filename, Lx = None, Ly = None, pos_x = None, pos_y = None, fix_aspect_ratio = False):
    r"""
    Read image using pyvista and interpolate on given mesh

    :param :class:`Mesh`: Target mesh
    :param str filename:  Filename of the image
    :param float Lx: Length to which the image should be scaled (defaults to the mesh length)
    :param float Ly: Height to which the image should be scaled (defaults to the mesh height)
    :param float pos_x: x-Offest by which the image should be shifted (defaults to the mesh origin)
    :param float pos_y: y-Offest by which the image should be shifted (defaults to the mesh origin)
    :param bool fix_aspect_ratio: if True only Lx or Ly can be set. The same scale will then be applied to both dimentions.
    :return :class:`torch.Tensor`: 2D tensor containing the correspoding image data

    :Examples:
      .. code::
        field = read_image(mesh, "measurement.png")
    """
    # read image data and convert to unique ids (0,1,2,...)
    image = pv.read(filename)
    data = image.get_array(image.array_names[0])
    if len(data.shape) == 2:
        data = np.prod(data, axis=1)
    data = np.unique(data, return_inverse=True)[1]
    data = data.reshape([image.dimensions[1], image.dimensions[0]]).T

    # scale and translate image
    if pos_x == None:
        pos_x = mesh.origin[0]
    if pos_y == None:
        pos_y = mesh.origin[1]

    if Lx != None and Ly != None and fix_aspect_ratio == True:
        raise RuntimeError("Aspect ratio cannot be kept fix, if both Lx and Ly are provided!")
    if Ly == None:
        Ly = mesh.n[1] * mesh.dx[1]
        if fix_aspect_ratio == True:
            Lx = Ly * image.dimensions[0] / image.dimensions[1]
    if Lx == None:
        Lx = mesh.n[0] * mesh.dx[0]
        if fix_aspect_ratio == True:
            Ly = Lx * image.dimensions[1] / image.dimensions[0]

    x_image = np.linspace(0,Lx,image.dimensions[0]) + pos_x
    y_image = np.linspace(0,Ly,image.dimensions[1]) + pos_y
    xx_image, yy_image = np.meshgrid(x_image, y_image, indexing = "ij")
    xx_image = xx_image.reshape(-1)
    yy_image = yy_image.reshape(-1)
    data = data.reshape(-1)

    # interpolate on mesh
    x = np.arange(mesh.n[0]) * mesh.dx[0] + mesh.dx[0]/2. + mesh.origin[0]
    y = np.arange(mesh.n[1]) * mesh.dx[1] + mesh.dx[1]/2. + mesh.origin[1]
    xx, yy = np.meshgrid(x, y, indexing = "ij")

    data = scipy.interpolate.griddata((xx_image, yy_image), data, (xx, yy), fill_value=-1)
    return torch.tensor(data)



def read_mesh(mesh, filename, scale = 1.):
    r"""
    Read unstructured msh meshes using pyvista

    :param str filename: Filename to be read
    :return :class:`Mesh` & dict: Mesh object and dictionary containing interpolated data tensors

    :Examples:
      .. code::
        fields = read_mesh(mesh, "cylinder.msh")
    """
    # read image data and volume domains
    unstructured_mesh = pv.read(filename)

    # interpolate on mesh
    x = np.arange(mesh.n[0]) * mesh.dx[0] + mesh.dx[0]/2. + mesh.origin[0]
    y = np.arange(mesh.n[1]) * mesh.dx[1] + mesh.dx[1]/2. + mesh.origin[1]
    z = np.arange(mesh.n[2]) * mesh.dx[2] + mesh.dx[2]/2. + mesh.origin[2]
    points = np.stack(np.meshgrid(x, y, z, indexing = "ij"), axis=-1).reshape(-1,3) / scale

    containing_cells = unstructured_mesh.find_containing_cell(points)
    data = unstructured_mesh.get_array(0)[containing_cells]
    data[containing_cells == -1] = -1 # containing_cell == -1, if point is not included in any cell

    return data.reshape(mesh.n)


def read_neper(filename, scale = 1.):
    r"""
    Read a rasterized Neper tessellation (.tesr) file.

    This function parses a Neper TESR file containing a rectilinear voxel
    representation of a tessellation. It extracts:

      • the grid dimensions and voxel spacings,
      • the per-cell (grain) IDs assigned to each voxel,
      • the per-cell group IDs (if present in the header).

    TESR files may contain both *id* and *group* lists in the header.
    The voxel data block stores only cell IDs, so group IDs are mapped
    voxel-wise using the cell → group association defined in the header.

    Parameters
    ----------
    filename : str
        Path to the `.tesr` file to be read.

    scale : float, optional
        Additional scaling factor applied to the voxel spacing
        (dx, dy, dz). Defaults to 1.

    Returns
    -------
    mesh : :class:`Mesh`
        A Mesh object describing grid resolution and voxel spacing.

    domains : torch.Tensor
        Integer tensor of shape (Nx, Ny, Nz) containing the **grain IDs**
        (cell IDs) for each voxel.

    groups : torch.Tensor
        Integer tensor of shape (Nx, Ny, Nz) containing the **group IDs**
        for each voxel. If the TESR file does not define groups, `groups`
        will contain zeros.

    Examples
    --------
    Generating a suitable Neper tessellation:

    .. code::

        # Create 1000 grains in a 500. x 250. x 40. domain
        # Assign group 1 to grains with id < 500, and group 2 otherwise
        # Rasterize to 500 x 125 x 10 voxels
        neper -T -n 1000 \
              -domain "cube(500,250,40)" \
              -group "id<500?1:2" \
              -o test \
              -format tess,tesr \
              -tesrsize 500:125:10

    Reading the generated TESR file:

    .. code::

        mesh, grain_ids, group_ids = read_neper("test.tesr")
    """
    Nx = Ny = Nz = None
    start_offset = None
    cell_ids = []
    group_ids = []

    with open(filename, "rb") as f:
        # --- Read ASCII header ---
        while True:
            pos = f.tell()
            line = f.readline().decode("ascii", errors="ignore")
            if not line:
                raise ValueError("Unexpected end of file while reading header.")

            line_s = line.strip()

            # --- Basic grid information ---
            if line_s.startswith("**general"):
                dim = int(f.readline().decode().strip())
                Nx, Ny, Nz = map(int, f.readline().decode().split())
                dx, dy, dz = tuple(scale * float(d) for d in f.readline().decode().split())

            # --- Cell definitions ---
            if line_s == "*id":
                # Next line(s) contain the cell IDs
                # They may span multiple lines until next "*"
                while True:
                    pos2 = f.tell()
                    l2 = f.readline().decode().strip()
                    if l2.startswith("*") or l2.startswith("**"):
                        f.seek(pos2)
                        break
                    cell_ids.extend(map(int, l2.split()))

            if line_s == "*group":
                # Next line(s) contain group IDs in same order as cell IDs
                while True:
                    pos2 = f.tell()
                    l2 = f.readline().decode().strip()
                    if l2.startswith("*") or l2.startswith("**"):
                        f.seek(pos2)
                        break
                    group_ids.extend(map(int, l2.split()))

            # --- Binary voxel block starts after **data / binary16 ---
            if line_s.startswith("**data"):
                fmt = f.readline().decode().strip()
                if fmt != "binary16":
                    raise ValueError(f"Unsupported TESR binary format: {fmt}")
                start_offset = f.tell()
                break

        if Nx is None:
            raise ValueError("TESR header did not contain voxel size.")

        # --- Read voxel cell ID block ---
        count = Nx * Ny * Nz
        vox_cell = np.fromfile(f, dtype="<u2", count=count)
        vox_cell = vox_cell.reshape((Nx, Ny, Nz), order="F")

    # --- Convert cell-id to group-id ---
    if group_ids:
        # cell_ids is 1-based, array index is 0-based
        lookup = np.zeros(max(cell_ids) + 1, dtype=np.int32)
        for cid, gid in zip(cell_ids, group_ids):
            lookup[cid] = gid
        vox_group = lookup[vox_cell]
    else:
        vox_group = None

    mesh = Mesh((Nx, Ny, Nz), (dx, dy, dz))
    domains = torch.tensor(vox_cell.astype(np.int64))
    groups = torch.tensor(vox_group.astype(np.int64))
    return mesh, domains, groups



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import torch 
import numpy as np 
from magnumnp.linear_elasticity.utils import gradient_with_pbc
from magnumnp.linear_elasticity.stress import sigma

__all__ = ["epsilon", "epsilon_m", "epsilon_el", "_get_C_jump_conditions", "_get_sigM_jump_conditions", "_get_B_jump_conditions"]

#@torch.compile
def epsilon(state, ud=None, iteration_depht = 1, second_order_boundary = False):
    if ud is None:
        ud = state.ud
    else:
        assert isinstance(ud, torch.Tensor) and ud.shape == state.mesh.n + (3,)

    mxl = torch.clone(state.m)
    mxr = torch.clone(state.m)
    myl = torch.clone(state.m)
    myr = torch.clone(state.m)
    mzl = torch.clone(state.m)
    mzr = torch.clone(state.m)

    dx_exp = state.mesh.dx_tensor[0].reshape(-1,1,1,1)
    dy_exp = state.mesh.dx_tensor[1].reshape(1,-1,1,1)
    dz_exp = state.mesh.dx_tensor[2].reshape(1,1,-1,1)

    A = state.material["A"][...,0]
    grad_mx = gradient_with_pbc(state.m[...,0], state.mesh, [0,1,2], [A, A, A], second_order_boundary=second_order_boundary)
    grad_my = gradient_with_pbc(state.m[...,1], state.mesh, [0,1,2], [A, A, A], second_order_boundary=second_order_boundary)
    grad_mz = gradient_with_pbc(state.m[...,2], state.mesh, [0,1,2], [A, A, A], second_order_boundary=second_order_boundary)

    n = state.mesh.n
    dmdx = torch.zeros(n+(3,))
    dmdx[...,0] = grad_mx[0]
    dmdx[...,1] = grad_my[0]
    dmdx[...,2] = grad_mz[0]

    dmdy = torch.zeros(n+(3,))
    dmdy[...,0] = grad_mx[1]
    dmdy[...,1] = grad_my[1]
    dmdy[...,2] = grad_mz[1]

    dmdz = torch.zeros(n+(3,))
    dmdz[...,0] = grad_mx[2]
    dmdz[...,1] = grad_my[2]
    dmdz[...,2] = grad_mz[2]

    mxl -= 0.5*dmdx*dx_exp
    mxr += 0.5*dmdx*dx_exp
    myl -= 0.5*dmdy*dy_exp
    myr += 0.5*dmdy*dy_exp
    mzl -= 0.5*dmdz*dz_exp
    mzr += 0.5*dmdz*dz_exp

    m_data = [mxl, mxr], [myl, myr], [mzl, mzr]
    C = _get_C_jump_conditions(state)
    Bl_sigM, Br_sigM = _get_sigM_jump_conditions(state, m_data)

    grad_ud_x = gradient_with_pbc(ud[...,0], state.mesh, [0,1,2], C[0], Bl_sigM[0], Br_sigM[0], second_order_boundary=second_order_boundary)
    grad_ud_y = gradient_with_pbc(ud[...,1], state.mesh, [0,1,2], C[1], Bl_sigM[1], Br_sigM[1], second_order_boundary=second_order_boundary)
    grad_ud_z = gradient_with_pbc(ud[...,2], state.mesh, [0,1,2], C[2], Bl_sigM[2], Br_sigM[2], second_order_boundary=second_order_boundary)

    if (iteration_depht > 0):
        for iter in range(iteration_depht):
            gradient_data = grad_ud_x, grad_ud_y, grad_ud_z

            Bl_eps, Br_eps = _get_B_jump_conditions(state, gradient_data)

            Bxl = [Bl_eps[0][0] + Bl_sigM[0][0], Bl_eps[0][1] + Bl_sigM[0][1], Bl_eps[0][2] + Bl_sigM[0][2]]
            Byl = [Bl_eps[1][0] + Bl_sigM[1][0], Bl_eps[1][1] + Bl_sigM[1][1], Bl_eps[1][2] + Bl_sigM[1][2]]
            Bzl = [Bl_eps[2][0] + Bl_sigM[2][0], Bl_eps[2][1] + Bl_sigM[2][1], Bl_eps[2][2] + Bl_sigM[2][2]]

            Bxr = [Br_eps[0][0] + Br_sigM[0][0], Br_eps[0][1] + Br_sigM[0][1], Br_eps[0][2] + Br_sigM[0][2]]
            Byr = [Br_eps[1][0] + Br_sigM[1][0], Br_eps[1][1] + Br_sigM[1][1], Br_eps[1][2] + Br_sigM[1][2]]
            Bzr = [Br_eps[2][0] + Br_sigM[2][0], Br_eps[2][1] + Br_sigM[2][1], Br_eps[2][2] + Br_sigM[2][2]]

            grad_ud_x[:] = gradient_with_pbc(ud[...,0], state.mesh, [0,1,2], C[0], Bxl, Bxr, second_order_boundary=second_order_boundary)[:]
            grad_ud_y[:] = gradient_with_pbc(ud[...,1], state.mesh, [0,1,2], C[1], Byl, Byr, second_order_boundary=second_order_boundary)[:]
            grad_ud_z[:] = gradient_with_pbc(ud[...,2], state.mesh, [0,1,2], C[2], Bzl, Bzr, second_order_boundary=second_order_boundary)[:]

    n = state.mesh.n
    eps = torch.zeros(n+(6,))
    eps[:,:,:,0] = grad_ud_x[0] # eps_0 = eps_xx
    eps[:,:,:,1] = grad_ud_y[1] # eps_1 = eps_yy
    eps[:,:,:,2] = grad_ud_z[2] # eps_2 = eps_zz
    eps[:,:,:,3] = grad_ud_y[2] + grad_ud_z[1] # eps_3 = 2*eps_yz
    eps[:,:,:,4] = grad_ud_x[2] + grad_ud_z[0] # eps_4 = 2*eps_xz
    eps[:,:,:,5] = grad_ud_x[1] + grad_ud_y[0] # eps_5 = 2*eps_xy

    return eps

@torch.compile
def epsilon_m(state, m=None):
    if m==None:
        m = state.m

    lambda_100 = state.material["lambda_100"][:,:,:,0]
    lambda_111 = state.material["lambda_111"][:,:,:,0]

    n = state.mesh.n
    eps_m = torch.zeros(n+(6,))

    eps_m[:,:,:,0] = (3./2.)*lambda_100*(m[:,:,:,0]**2. - 1./3.)
    eps_m[:,:,:,1] = (3./2.)*lambda_100*(m[:,:,:,1]**2. - 1./3.)
    eps_m[:,:,:,2] = (3./2.)*lambda_100*(m[:,:,:,2]**2. - 1./3.)
    eps_m[:,:,:,3] = 3.*lambda_111*m[:,:,:,1]*m[:,:,:,2]
    eps_m[:,:,:,4] = 3.*lambda_111*m[:,:,:,0]*m[:,:,:,2]
    eps_m[:,:,:,5] = 3.*lambda_111*m[:,:,:,0]*m[:,:,:,1]

    return eps_m

def epsilon_el(state, ud=None, m=None):
    return epsilon(state, ud) - epsilon_m(state, m)

def _get_C_jump_conditions(state):
    C = state.material["C"]
    C11 = C[...,0,0]
    C22 = C[...,1,1]
    C33 = C[...,2,2]
    C44 = C[...,3,3]
    C55 = C[...,4,4]
    C66 = C[...,5,5]

    # for x derivatives
    Cxx = C11
    Cyx = C66
    Czx = C55

    # for y derivatives
    Cxy = C66
    Cyy = C22
    Czy = C44

    # for z derivatives
    Cxz = C55
    Cyz = C44
    Czz = C33

    return [Cxx, Cxy, Cxz], [Cyx, Cyy, Cyz], [Czx, Czy, Czz]

def _get_sigM_jump_conditions(state, m_data):
    mxl, mxr = m_data[0]
    myl, myr = m_data[1]
    mzl, mzr = m_data[2]
    
    # for x derivatives
    eps_m = epsilon_m(state, mxl)
    sig_m = sigma(state,eps_m)
    Bx_xl = -sig_m[...,0]
    By_xl = -sig_m[...,5]
    Bz_xl = -sig_m[...,4]

    eps_m[:] = epsilon_m(state, mxr)
    sig_m[:] = sigma(state,eps_m)
    Bx_xr = -sig_m[...,0]
    By_xr = -sig_m[...,5]
    Bz_xr = -sig_m[...,4]

    # for y derivatives
    eps_m[:] = epsilon_m(state, myl)
    sig_m[:] = sigma(state,eps_m)
    Bx_yl = -sig_m[...,5]
    By_yl = -sig_m[...,1]
    Bz_yl = -sig_m[...,4]

    eps_m[:] = epsilon_m(state, myr)
    sig_m[:] = sigma(state,eps_m)
    Bx_yr = -sig_m[...,5]
    By_yr = -sig_m[...,1]
    Bz_yr = -sig_m[...,4]

    # for z derivatives
    eps_m[:] = epsilon_m(state, mzl)
    sig_m[:] = sigma(state,eps_m)
    Bx_zl = -sig_m[...,4]
    By_zl = -sig_m[...,3]
    Bz_zl = -sig_m[...,2]

    eps_m[:] = epsilon_m(state, mzr)
    sig_m[:] = sigma(state,eps_m)
    Bx_zr = -sig_m[...,4]
    By_zr = -sig_m[...,3]
    Bz_zr = -sig_m[...,2]

    Bl = [[Bx_xl, Bx_yl, Bx_zl], [By_xl, By_yl, By_zl], [Bz_xl, Bz_yl, Bz_zl]]
    Br = [[Bx_xr, Bx_yr, Bx_zr], [By_xr, By_yr, By_zr], [Bz_xr, Bz_yr, Bz_zr]]

    return Bl, Br

def _get_B_jump_conditions(state, gradient_data):
    gux, guy, guz = gradient_data

    C = state.material["C"]
    C12 = C[...,0,1]
    C13 = C[...,0,2]
    C23 = C[...,1,2]
    C44 = C[...,3,3]
    C55 = C[...,4,4]
    C66 = C[...,5,5]

    def harmonic_mean(g, C, shift, dim):
        a = C*g
        mean = (a+torch.roll(a, shift, dim)) / (C + torch.roll(C, shift, dim))
        return mean.nan_to_num(posinf=0, neginf=0)
    
    dyux_xl = harmonic_mean(gux[1], C66, 1, 0)
    dzux_xl = harmonic_mean(gux[2], C55, 1, 0)
    dxux_yl = harmonic_mean(gux[0], C12, 1, 1)
    dxux_zl = harmonic_mean(gux[0], C13, 1, 2)

    dxuy_yl = harmonic_mean(guy[0], C66, 1, 1)
    dzuy_yl = harmonic_mean(guy[2], C44, 1, 1)
    dyuy_xl = harmonic_mean(guy[1], C12, 1, 0)
    dyuy_zl = harmonic_mean(guy[1], C23, 1, 2)

    dxuz_zl = harmonic_mean(guz[0], C55, 1, 2)
    dyuz_zl = harmonic_mean(guz[1], C44, 1, 2)
    dzuz_xl = harmonic_mean(guz[2], C13, 1, 0)
    dzuz_yl = harmonic_mean(guz[2], C23, 1, 1)

    dyux_xr = torch.roll(dyux_xl, -1, 0)
    dzux_xr = torch.roll(dzux_xl, -1, 0)
    dxux_yr = torch.roll(dxux_yl, -1, 1)
    dxux_zr = torch.roll(dxux_zl, -1, 2)

    dxuy_yr = torch.roll(dxuy_yl, -1, 1)
    dzuy_yr = torch.roll(dzuy_yl, -1, 1)
    dyuy_xr = torch.roll(dyuy_xl, -1, 0)
    dyuy_zr = torch.roll(dyuy_zl, -1, 2)

    dxuz_zr = torch.roll(dxuz_zl, -1, 2)
    dyuz_zr = torch.roll(dyuz_zl, -1, 2)
    dzuz_xr = torch.roll(dzuz_xl, -1, 0)
    dzuz_yr = torch.roll(dzuz_yl, -1, 1)

    # for x derivatives
    Bx_xl = C12*dyuy_xl + C13*dzuz_xl
    By_xl = C66*dyux_xl
    Bz_xl = C55*dzux_xl

    Bx_xr = C12*dyuy_xr + C13*dzuz_xr
    By_xr = C66*dyux_xr
    Bz_xr = C55*dzux_xr

    # for y derivatives
    Bx_yl = C66*dxuy_yl
    By_yl = C12*dxux_yl + C23*dzuz_yl
    Bz_yl = C44*dzuy_yl

    Bx_yr = C66*dxuy_yr
    By_yr = C12*dxux_yr + C23*dzuz_yr
    Bz_yr = C44*dzuy_yr

    # for z derivatives
    Bx_zl = C55*dxuz_zl
    By_zl = C44*dyuz_zl
    Bz_zl = C13*dxux_zl + C23*dyuy_zl

    Bx_zr = C55*dxuz_zr
    By_zr = C44*dyuz_zr
    Bz_zr = C13*dxux_zr + C23*dyuy_zr

    Bl = [[Bx_xl, Bx_yl, Bx_zl], [By_xl, By_yl, By_zl], [Bz_xl, Bz_yl, Bz_zl]]
    Br = [[Bx_xr, Bx_yr, Bx_zr], [By_xr, By_yr, By_zr], [Bz_xr, Bz_yr, Bz_zr]]
    
    return Bl, Br


#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from magnumnp.common import constants
from magnumnp.utils.logging_helpers import LogDt
from magnumnp.field_terms import OerstedField, VectorPotential

__all__ = ["Coil"]

class Coil(object):
    def __init__(self, state, I0, h_oersted0 = None, A0 = None):
        self.I0 = I0

        self.h_oersted0 = h_oersted0
        if self.h_oersted0 == None:
            self.h_oersted0 = OerstedField().h(state)

        self.A0 = A0
        if A0 == None:
            self.A0 = VectorPotential().A(state)

        # calculate induced voltage by numerical time-derivative
        self.Ui_m = LogDt(self.psi_m)
        self.Ui_j = LogDt(self.psi_j)
        self.Ui = LogDt(self.psi)

        # calculate stationary parameters
        self.L = self.psi_j(state) / self.I0
        self.R = (state.j**2 / state.material["sigma"]).nan_to_num(posinf=0, neginf=0).sum() * state.mesh.cell_volumes / self.I0**2

    def psi_m(self, state):
        return (state.material["Ms"] * state.m * self.h_oersted0).sum() / self.I0 * constants.mu_0 * state.mesh.cell_volumes

    def psi_j(self, state):
        return (self.A0*state.j).sum() * constants.mu_0 * state.mesh.cell_volumes

    def psi(self, state):
        return self.psi_m(state) + self.psi_j(state)


# state.j = ...
# h_oersted = OerstedField().h(state)
# coil = Coil(state, I0=j0*A)
#
# logger = Logger("data", ["t", "m", coil.U_m, coil.U_j, coil.U])




#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import logging
import os
import sys

__all__ = ["logger", "set_log_level", "set_log_file", "set_log_script", "debug", "warning", "error", "info", "default", "info_green", "info_blue"]

INFO_GREEN = logging.INFO+5
INFO_BLUE = logging.INFO
DEFAULT = logging.CRITICAL
logging.addLevelName(INFO_GREEN, "INFO")
logging.addLevelName(INFO_BLUE, "INFO")
logging.addLevelName(DEFAULT, "")

# create magnum.fe logger
logger = logging.getLogger('magnum.np')

handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter(fmt="%(asctime)s  %(name)s:%(levelname)s %(message)s", datefmt='%Y-%m-%d %H:%M:%S'))
logger.addHandler(handler)

logger.setLevel(logging.INFO)

info = logger.info

RED = "\033[1;37;31m%s\033[0m"
BLUE = "\033[1;37;34m%s\033[0m"
GREEN = "\033[1;37;32m%s\033[0m"
CYAN = "\033[1;37;36m%s\033[0m"

def default(message, *args, **kwargs):
    logger.log(DEFAULT, message, *args, **kwargs)

def debug(message, *args, **kwargs):
    logger.debug(CYAN % message, *args, **kwargs)

def warning(message, *args, **kwargs):
    logger.warning(RED % message, *args, **kwargs)

def error(message, *args, **kwargs):
    logger.error(RED % message, *args, **kwargs)

def info_green(message, *args, **kwargs):
    logger.log(INFO_GREEN, GREEN % message, *args, **kwargs)

def info_blue(message, *args, **kwargs):
    logger.log(INFO_BLUE, BLUE % message, *args, **kwargs)

def set_log_level(level):
    """
    Set the log level of magnum.np specific logging messages.
    Defaults to :code:`INFO = 20`.

    *Arguments*
      level (:class:`int`)
        The log level
    """
    logger.setLevel(level)

def set_log_file(filename):
    """
    Store logging output to specified file.

    *Arguments*
        filename (:class:`str`)
    """
    # create directory if not existent
    if not os.path.dirname(filename) == '' and \
         not os.path.exists(os.path.dirname(filename)):
        try:
            os.makedirs(os.path.dirname(filename))
        except OSError as exc: # Guard against race condition
            if exc.errno != errno.EEXIST:
                raise

    handler = logging.FileHandler(filename, mode='w')
    handler.setFormatter(logging.Formatter(fmt="%(asctime)s  %(name)s:%(levelname)s %(message)s", datefmt='%Y-%m-%d %H:%M:%S'))
    logger.addHandler(handler)

def set_log_script(filename):
    """
    Copy run script to specified path

    *Arguments*
        filename (:class:`str`)
    """
    # create directory if not existent
    if not os.path.dirname(filename) == '' and \
         not os.path.exists(os.path.dirname(filename)):
        try:
            os.makedirs(os.path.dirname(filename))
        except OSError as exc: # Guard against race condition
            if exc.errno != errno.EEXIST:
                raise
    open(filename, 'wb').write(open(sys.argv[0], 'rb').read())



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from magnumnp.common import timedmethod, constants
from magnumnp.field_terms import LinearFieldTerm
import torch

__all__ = ["IntergrainExchangeField", "RKKYField", "BiquadraticRKKYField"]

class IntergrainExchangeField(LinearFieldTerm):
    r"""
    Intergrain - Exchange interaction between two domains gives rise to the following energy contribution:

    .. math::

        E^\text{iex} = -\int\limits_\Gamma J_\text{iex} \, \vec{m}_i \cdot \vec{m}_j \, d\vec{A},

    where :math:`\Gamma` is the interface between two layers :math:`i` and :math:`j`
    with magnetizations :math:`\vec{m}_i` and :math:`\vec{m}_j`, respectively.

    :Example:

      .. code::

        # create state with named domains from mesh
        state = State(mesh)
        state.material["iex"] = J_iex

        # create domains as bool arrays, e.g:
        domain1 = torch.zeros(n, dtype=torch.bool)
        domain1[n[0]//2:,:,:] = True

        domain2 = torch.zeros(n, dtype=torch.bool)
        domain2[:-n[0]//2:,:,:] = True

        # rotate magnetization within one subdomain
        state.m[domain1] = torch.tensor([np.cos(phi), np.sin(phi), 0])

        # without interface layer, two seperate exchange fields need to be defined
        exchange1 = ExchangeField(Aex1, domain1)
        exchange2 = ExchangeField(Aex2, domain2)
        iex = IntergrainExchangeField(domain1, domain2)
    """
    def __init__(self, domain1, domain2):
        self._domain1 = domain1
        self._domain2 = domain2

    @timedmethod
    @torch.compile
    def h(self, state):
        m1 = state.m * state.material["iex"]
        m2 = m1.clone()
        m1[~self._domain1] = 0
        m2[~self._domain2] = 0

        # sum h1 over all next neighbors
        h1 = m1*0.
        h1[:-1,:,:,:] += m2[ 1:,:,:,:] / state.mesh.dx[0]
        h1[ 1:,:,:,:] += m2[:-1,:,:,:] / state.mesh.dx[0]

        h1[:,:-1,:,:] += m2[:, 1:,:,:] / state.mesh.dx[1]
        h1[:, 1:,:,:] += m2[:,:-1,:,:] / state.mesh.dx[1]

        h1[:,:,:-1,:] += m2[:,:, 1:,:] / state.mesh.dx[2]
        h1[:,:, 1:,:] += m2[:,:,:-1,:] / state.mesh.dx[2]

        # sum h2 over all next neighbors
        h2 = m2*0.
        h2[:-1,:,:,:] += m1[ 1:,:,:,:] / state.mesh.dx[0]
        h2[ 1:,:,:,:] += m1[:-1,:,:,:] / state.mesh.dx[0]

        h2[:,:-1,:,:] += m1[:, 1:,:,:] / state.mesh.dx[1]
        h2[:, 1:,:,:] += m1[:,:-1,:,:] / state.mesh.dx[1]

        h2[:,:,:-1,:] += m1[:,:, 1:,:] / state.mesh.dx[2]
        h2[:,:, 1:,:] += m1[:,:,:-1,:] / state.mesh.dx[2]

        h = m1 * 0.
        h[self._domain1] = h1[self._domain1]
        h[self._domain2] = h2[self._domain2]

        h /= constants.mu_0 * state.material["Ms"]
        return h.nan_to_num(posinf=0, neginf=0)




# TODO: interface should be generalized and simplified
class RKKYField(object):
    r"""
    Interlayer-Exchange interaction between two layers gives rise to the following energy contribution:

    .. math::

        E^\text{rkky} = -\int\limits_\Gamma J_\text{rkky} \, \vec{m}_i \cdot \vec{m}_j \, d\vec{A},

    where :math:`\Gamma` is the interface between two layers :math:`i` and :math:`j`
    with magnetizations :math:`\vec{m}_i` and :math:`\vec{m}_j`, respectively.

    Special care has to be taken, if domain walls or partial domain walls are formed across the RKKY interface.
    In this case higher order approximations of the magnetization needs to be used near the interface in order to
    accurately describe e.g. the equilibrium magnetization or critical switching fields.
    (see Suess et al. "Accurate finite difference micromagnetic of magnetis including RKKY interaction -- analytical solution and comparision to standard micromagnetic codes")

    :param J_rkky: Interlayer-Exchange constant :math:`J_\text{rkky}`
    :type J_rkky: float
    :param dir: normal direction of the interface (currently "z" is hard-coded")
    :type filename: str
    :param id1: Index of the first layer
    :type id1: int
    :param id2: Index of the second layer
    :type id2: int
    :param order: appoximation order of the magnetization near the interface (default = 0)
    :type order: int, optional

    :Example:

      .. code::

        # create state with named domains from mesh
        state = State(mesh)

        # create domains as bool arrays, e.g:
        domain1 = torch.zeros(n, dtype=torch.bool)
        domain1[n[0]//2:,:,:] = True

        domain2 = torch.zeros(n, dtype=torch.bool)
        domain2[:-n[0]//2:,:,:] = True

        # rotate magnetization within one subdomain
        state.m[domain1] = torch.tensor([np.cos(phi), np.sin(phi), 0])

        # without interface layer, two seperate exchange fields need to be defined
        exchange1 = ExchangeField(Aex1, domain1)
        exchange2 = ExchangeField(Aex2, domain2)
        rkky = RKKYField(J_rkky, "z", id1, id2)
    """
    def __init__(self, J_rkky, dir, id1, id2, order = 0):
        self._J_rkky = J_rkky
        if dir != "z":
            raise ValueError("Currently only dir='z' is implemented!")
        self._dir = dir #TODO: dir is ignored
        self._id1 = min(id1,id2)
        self._id2 = max(id1,id2)
        self._order = order

    @timedmethod
    @torch.compile
    def h(self, state):
        h = torch.zeros_like(state.m)
        dz = state.mesh.dx_tensor[2].reshape(1,1,-1,1)
        if self._order != 0 and not state.mesh.is_equidistant:
            raise ValueError("[RKKYField] higher order RKKY is only implemented for equidistant fields! (use order=0)")
        if self._order == 0:
            m1 = state.m[:,:,(self._id1,),:]
            m2 = state.m[:,:,(self._id2,),:]
            h[:,:,(self._id1,),:] = self._J_rkky * m2 # this form makes the 0-order field linear
            h[:,:,(self._id2,),:] = self._J_rkky * m1 # (which is needed for eigenmode calculations)
        else:
            if self._order == 1:
                m1 = 1.5 * state.m[:,:,(self._id1,),:] - 0.5 * state.m[:,:,(self._id1-1,),:]
                m2 = 1.5 * state.m[:,:,(self._id2,),:] - 0.5 * state.m[:,:,(self._id2+1,),:]
            elif self._order == 2:
                m1 = 15./8. * state.m[:,:,(self._id1,),:] - 5./4. * state.m[:,:,(self._id1-1,),:] + 3./8.* state.m[:,:,(self._id1-2,),:]
                m2 = 15./8. * state.m[:,:,(self._id2,),:] - 5./4. * state.m[:,:,(self._id2+1,),:] + 3./8.* state.m[:,:,(self._id2+2,),:]
            h[:,:,(self._id1,),:] = self._J_rkky * (m2 - (m1*m2).sum(axis = 3, keepdim=True) * m1)
            h[:,:,(self._id2,),:] = self._J_rkky * (m1 - (m1*m2).sum(axis = 3, keepdim=True) * m2)

        h /= constants.mu_0 * state.material["Ms"] * dz
        return h.nan_to_num(posinf=0, neginf=0)

    @torch.compile
    def E(self, state):
        m1 = state.m[:,:,(self._id1,),:]
        m2 = state.m[:,:,(self._id2,),:]
        dx = state.mesh.dx_tensor[0].reshape(-1,1,1,1)
        dy = state.mesh.dx_tensor[1].reshape(1,-1,1,1)
        if self._order == 1:
            m1 += 0.5 * ( state.m[:,:,(self._id1,),:] - state.m[:,:,(self._id1-1,),:])
            m2 += 0.5 * (-state.m[:,:,(self._id2,),:] + state.m[:,:,(self._id2+1,),:])
        elif self._order == 2:
            m1 += 0.25 * (3*state.m[:,:,(self._id1,),:] - 4*state.m[:,:,(self._id1-1,),:] + state.m[:,:,(self._id1-2,),:])
            m2 += 0.25 * (3*state.m[:,:,(self._id2,),:] - 4*state.m[:,:,(self._id2+1,),:] + state.m[:,:,(self._id2+2,),:])

        E = -(dx * dy * self._J_rkky * m1 * m2).sum()
        return E


# TODO: interface should be generalized and simplified
class BiquadraticRKKYField(object):
    r"""
    Biquadratic surface exchange couplong between two layers gives rise to the following energy contribution:

    .. math::

        E^\text{biquadratic} = -\int\limits_\Gamma J_\text{biquadratic} \, (\vec{m}_i \cdot \vec{m}_j)^2 \, d\vec{A},

    where :math:`\Gamma` is the interface between two layers :math:`i` and :math:`j` with magnetizations :math:`\vec{m}_i` and :math:`\vec{m}_j`, respectively.

    The effective field is given by:

    .. math::

        \vec{h}^\text{biquadratic}_i = \frac{2 J_\text{biquadratic}} {M_s \Delta z \mu_0} \, (\vec{m}_i \cdot \vec{m}_j) \, \vec{m}_j,

    with the interlayer exchange constant :math:`J_\text{biquadratic}`.
    """
    def __init__(self, J_rkky_BQ, dir, id1, id2):
        self._J_rkky_BQ = J_rkky_BQ
        if dir != "z":
            raise ValueError("Currently only dir='z' is implemented!")
        self._dir = dir
        self._id1 = min(id1,id2)
        self._id2 = max(id1,id2)

    @timedmethod
    @torch.compile
    def h(self, state):
        h = torch.zeros(state.mesh.n + (3,))
        dz = state.mesh.dx_tensor[2].reshape(1,1,-1,1)

        m1 = state.m[:,:,self._id1,:]
        m2 = state.m[:,:,self._id2,:]

        m12 = (m1*m2).sum(axis=-1, keepdim=True)
        h[:,:,self._id1,:] = 2. * self._J_rkky_BQ * m12 * m2
        h[:,:,self._id2,:] = 2. * self._J_rkky_BQ * m12 * m1

        #TODO: find out why there is a 2x discrepancy compared with oommf
        h /= constants.mu_0 * state.material["Ms"] * dz

        return h.nan_to_num(posinf=0, neginf=0)

    @torch.compile
    def E(self, state):
        m1 = state.m[:,:,self._id1,:]
        m2 = state.m[:,:,self._id2,:]
        dx = state.mesh.dx_tensor[0].reshape(-1,1,1)
        dy = state.mesh.dx_tensor[1].reshape(1,-1,1)

        E = -dx * dy * self._J_rkky_BQ * (m1*m2).sum(dim=-1,keepdim=True)**2
        return E.sum()



#
# This file is part of the magnum.np distribution
# (https://gitlab.com/magnum.np/magnum.np).
# Copyright (c) 2023 magnum.np team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import torch
from magnumnp.common import logging
from torchdiffeq import odeint, odeint_adjoint

__all__ = ["TorchDiffEq", "TorchDiffEqAdjoint"]

class TorchDiffEq(object):
    def __init__(self, f, method = "dopri5", rtol = 1e-5, atol = 1e-5, options = {}):
        self._f = f
        self._method = method
        self._rtol = rtol
        self._atol = atol
        self._options = options
        logging.info_green("[LLGSolver] using TorchDiffEq solver (method = '%s', rtol = %g, atol = %g)" % (method, rtol, atol))

    def _f_wrapper(self, t, x, **kwargs): # TODO: move scaling to dm?
        return self._f(t * 1e-9, x, **kwargs) * 1e-9 # scale time by 1e9 to prevent underflow error

    def step(self, t, x, dt, rtol = None, atol = None, **kwargs):
        t1 = t + dt
        res = odeint(lambda t, x: self._f_wrapper(t, x, **kwargs),
                     x,
                     torch.tensor([t*1e9, t1*1e9]),
                     method = self._method,
                     rtol = rtol or self._rtol,
                     atol = atol or self._atol,
                     options = self._options) # TODO: reuse solver object?
        return t1, res[1]

    def solve(self, tt, x, rtol = None, atol = None, **kwargs):
        res = odeint(lambda t, x: self._f_wrapper(t, x, **kwargs),
                     x,
                     tt*1e9,
                     method = self._method,
                     rtol = rtol or self._rtol,
                     atol = atol or self._atol,
                     options = self._options)
        return res



class TorchDiffEqAdjoint(object):
    def __init__(self, f, adjoint_parameters, method = "dopri5", rtol = 1e-5, atol = 1e-5, options = {}):
        self._f = f
        if not (isinstance(adjoint_parameters, list) or isinstance(adjoint_parameters, tuple)):
            raise ValueError("[TorchDiffEqAdjoint] adjoint_parameters must be a List (use [adjoint_parameters]).")
        self._adjoint_parameters = adjoint_parameters
        self._method = method
        self._rtol = rtol
        self._atol = atol
        self._options = options
        logging.info_green("[LLGSolver] using TorchDiffEq adjoint solver (method = '%s', rtol = %g, atol = %g)" % (method, rtol, atol))

    def _f_wrapper(self, t, x, **kwargs): # TODO: move scaling to dm?
        return self._f(t * 1e-9, x, **kwargs) * 1e-9 # scale time by 1e9 to prevent underflow error

    def step(self, t, x, dt, rtol = None, atol = None, **kwargs):
        t1 = t + dt
        res = odeint_adjoint(lambda t, x: self._f_wrapper(t, x, **kwargs),
                     x,
                     torch.tensor([t*1e9, t1*1e9]),
                     method = self._method,
                     rtol = rtol or self._rtol,
                     atol = atol or self._atol,
                     adjoint_params = self._adjoint_parameters,
                     options = self._options) # TODO: reuse solver object?
        return t1, res[1]

    def solve(self, tt, x, rtol = None, atol = None, **kwargs):
        res = odeint_adjoint(lambda t, x: self._f_wrapper(t, x, **kwargs),
                     x,
                     tt*1e9,
                     method = self._method,
                     rtol = rtol or self._rtol,
                     atol = atol or self._atol,
                     adjoint_params = self._adjoint_parameters,
                     options = self._options)
        return res
